from __future__ import annotations

from functools import lru_cache
from inspect import isawaitable
from typing import TYPE_CHECKING, Any, Callable

from graphql import GraphQLSchema, execute, parse, validate
from graphql.error import GraphQLError
from graphql.execution import ExecutionResult
from graphql.type import validate_schema

if TYPE_CHECKING:
    from graphql import ExecutionContext, GraphQLFieldResolver, GraphQLTypeResolver
    from graphql.execution import Middleware
    from graphql.language import Source
    from graphql.language.ast import DocumentNode


@lru_cache(maxsize=1024)
def _cached_parse(query: str) -> DocumentNode:
    """Internal cached parse function for queries without @expand directive."""
    return parse(query)


def cached_parse(query: str | Source) -> DocumentNode:
    """Parse a GraphQL query string into a DocumentNode.

    Queries containing the @expand directive are not cached because the parser
    mutates the AST to add expanded fields, which would corrupt the cache.
    """
    query_str = query if isinstance(query, str) else query.body
    if "@expand" in query_str:
        return parse(query)
    return _cached_parse(query_str)


@lru_cache(maxsize=1024)
def cached_validate(schema: GraphQLSchema, document_ast: DocumentNode) -> list[GraphQLError]:
    return validate(schema, document_ast)


@lru_cache(maxsize=1024)
def cached_validate_schema(schema: GraphQLSchema) -> list[GraphQLError]:
    return validate_schema(schema)


async def execute_graphql_query(
    schema: GraphQLSchema,
    source: str | Source,
    root_value: Any = None,
    context_value: Any = None,
    variable_values: dict[str, Any] | None = None,
    operation_name: str | None = None,
    field_resolver: GraphQLFieldResolver | None = None,
    type_resolver: GraphQLTypeResolver | None = None,
    middleware: Middleware | None = None,
    execution_context_class: type[ExecutionContext] | None = None,
    is_awaitable: Callable[[Any], bool] | None = None,
) -> ExecutionResult:
    """Execute a query, return asynchronously only if necessary."""
    # Validate Schema
    schema_validation_errors = cached_validate_schema(schema)
    if schema_validation_errors:
        return ExecutionResult(data=None, errors=schema_validation_errors)

    # Parse
    try:
        document = cached_parse(source)
    except GraphQLError as error:
        return ExecutionResult(data=None, errors=[error])

    validation_errors = cached_validate(schema, document)
    if validation_errors:
        return ExecutionResult(data=None, errors=validation_errors)

    # Execute
    result = execute(
        schema,
        document,
        root_value,
        context_value,
        variable_values,
        operation_name,
        field_resolver,
        type_resolver,
        None,
        middleware,
        execution_context_class,
        is_awaitable,
    )

    if isawaitable(result):
        return await result

    return result



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from rich.progress import Progress

from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.constants import BranchSupportType, SchemaPathType
from infrahub.core.initialization import get_root_node
from infrahub.core.migrations.schema.node_attribute_add import NodeAttributeAddMigration
from infrahub.core.migrations.shared import (
    MigrationInput,
    MigrationRequiringRebase,
    MigrationResult,
    get_migration_console,
)
from infrahub.core.path import SchemaPath
from infrahub.core.query import Query, QueryType

from .load_schema_branch import get_or_load_schema_branch

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class GetAddedNodesByKindForBranchQuery(Query):
    name = "get_added_nodes_by_kind_for_branch_query"
    type = QueryType.READ
    insert_return = True

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["branch"] = self.branch.name
        query = """
MATCH (n:Node)-[e:IS_PART_OF {branch: $branch, status: "active"}]->(:Root)
WHERE e.to IS NULL
AND NOT exists((n)-[:IS_PART_OF {branch: $branch, status: "deleted"}]->(:Root))
WITH n.kind AS kind, collect(n.uuid) AS node_ids
        """
        self.return_labels = ["kind", "node_ids"]
        self.add_to_query(query)

    def get_node_ids_by_kind(self) -> dict[str, list[str]]:
        node_ids_by_kind: dict[str, list[str]] = {}
        for result in self.get_results():
            kind = result.get_as_type(label="kind", return_type=str)
            node_ids: list[str] = result.get_as_type(label="node_ids", return_type=list)
            node_ids_by_kind[kind] = node_ids
        return node_ids_by_kind


class Migration043(MigrationRequiringRebase):
    name: str = "043_create_hfid_display_label_in_db"
    minimum_version: int = 42

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        result = MigrationResult()

        root_node = await get_root_node(db=db, initialize=False)
        default_branch_name = root_node.default_branch
        default_branch = await Branch.get_by_name(db=db, name=default_branch_name)
        main_schema_branch = await get_or_load_schema_branch(db=db, branch=default_branch)
        schema_node = main_schema_branch.get_node(name="SchemaNode")
        schema_generic = main_schema_branch.get_node(name="SchemaGeneric")

        migrations = [
            # HFID is not needed, it was introduced at graph v8
            NodeAttributeAddMigration(
                new_node_schema=schema_node,
                previous_node_schema=schema_node,
                schema_path=SchemaPath(
                    schema_kind="SchemaNode", path_type=SchemaPathType.ATTRIBUTE, field_name="display_label"
                ),
            ),
            NodeAttributeAddMigration(
                new_node_schema=schema_generic,
                previous_node_schema=schema_generic,
                schema_path=SchemaPath(
                    schema_kind="SchemaGeneric", path_type=SchemaPathType.ATTRIBUTE, field_name="display_label"
                ),
            ),
        ]

        for node_schema_kind in main_schema_branch.node_names:
            schema = main_schema_branch.get(name=node_schema_kind, duplicate=False)
            if schema.branch is not BranchSupportType.AWARE:
                continue
            migrations.extend(
                [
                    NodeAttributeAddMigration(
                        new_node_schema=schema,
                        previous_node_schema=schema,
                        schema_path=SchemaPath(
                            schema_kind=schema.kind, path_type=SchemaPathType.ATTRIBUTE, field_name="human_friendly_id"
                        ),
                    ),
                    NodeAttributeAddMigration(
                        new_node_schema=schema,
                        previous_node_schema=schema,
                        schema_path=SchemaPath(
                            schema_kind=schema.kind, path_type=SchemaPathType.ATTRIBUTE, field_name="display_label"
                        ),
                    ),
                ]
            )

        with Progress(console=get_migration_console()) as progress:
            update_task = progress.add_task("Adding HFID and display label to nodes", total=len(migrations))

            for migration in migrations:
                try:
                    execution_result = await migration.execute(migration_input=migration_input, branch=default_branch)
                    result.errors.extend(execution_result.errors)
                    progress.update(update_task, advance=1)
                except Exception as exc:
                    result.errors.append(str(exc))
                    return result

        return result

    async def execute_against_branch(self, migration_input: MigrationInput, branch: Branch) -> MigrationResult:
        db = migration_input.db
        result = MigrationResult()

        schema_branch = await registry.schema.load_schema_from_db(db=db, branch=branch)

        migrations = []
        get_added_nodes_by_kind_for_branch_query = await GetAddedNodesByKindForBranchQuery.init(db=db, branch=branch)
        await get_added_nodes_by_kind_for_branch_query.execute(db=db)
        node_ids_by_kind = get_added_nodes_by_kind_for_branch_query.get_node_ids_by_kind()

        for node_kind, node_ids in node_ids_by_kind.items():
            schema = schema_branch.get(name=node_kind, duplicate=False)
            if schema.branch not in (BranchSupportType.AWARE, BranchSupportType.LOCAL):
                continue
            migrations.extend(
                [
                    NodeAttributeAddMigration(
                        uuids=node_ids,
                        new_node_schema=schema,
                        previous_node_schema=schema,
                        schema_path=SchemaPath(
                            schema_kind=schema.kind, path_type=SchemaPathType.ATTRIBUTE, field_name="human_friendly_id"
                        ),
                    ),
                    NodeAttributeAddMigration(
                        uuids=node_ids,
                        new_node_schema=schema,
                        previous_node_schema=schema,
                        schema_path=SchemaPath(
                            schema_kind=schema.kind, path_type=SchemaPathType.ATTRIBUTE, field_name="display_label"
                        ),
                    ),
                ]
            )

        with Progress(console=get_migration_console()) as progress:
            update_task = progress.add_task(
                f"Adding HFID and display label to nodes on branch {branch.name}", total=len(migrations)
            )

            for migration in migrations:
                try:
                    execution_result = await migration.execute(migration_input=migration_input, branch=branch)
                    result.errors.extend(execution_result.errors)
                    progress.update(update_task, advance=1)
                except Exception as exc:
                    result.errors.append(str(exc))
                    return result

        return result

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()



import asyncio
import contextlib
import logging
import os
from pathlib import Path
from typing import Any

import typer
from anyio.abc import TaskStatus
from infrahub_sdk import Config, InfrahubClient
from infrahub_sdk.exceptions import Error as SdkError
from prefect import settings as prefect_settings
from prefect.client.schemas.objects import FlowRun
from prefect.context import AsyncClientContext
from prefect.flow_engine import run_flow_async
from prefect.logging.handlers import APILogHandler
from prefect.workers.base import BaseJobConfiguration, BaseVariables, BaseWorker, BaseWorkerResult
from prometheus_client import start_http_server

from infrahub import __version__ as infrahub_version
from infrahub import config
from infrahub.components import ComponentType
from infrahub.core import registry
from infrahub.core.initialization import initialization
from infrahub.database.graph import validate_graph_version
from infrahub.dependencies.registry import build_component_registry
from infrahub.exceptions import InitializationError
from infrahub.git import initialize_repositories_directory
from infrahub.lock import initialize_lock
from infrahub.services import InfrahubServices
from infrahub.trace import configure_trace
from infrahub.workers.dependencies import (
    get_cache,
    get_component,
    get_database,
    get_http,
    get_message_bus,
    get_workflow,
    set_component_type,
)
from infrahub.workers.utils import inject_service_parameter, load_flow_function
from infrahub.workflows.models import TASK_RESULT_STORAGE_NAME

WORKER_QUERY_SECONDS = "2"
WORKER_DEFAULT_RESULT_STORAGE_BLOCK = f"redisstoragecontainer/{TASK_RESULT_STORAGE_NAME}"
DEFAULT_TASK_LOGGERS = ["infrahub.tasks"]


class InfrahubWorkerAsyncConfiguration(BaseJobConfiguration):
    env: dict[str, str | None] = {
        "PREFECT_WORKER_QUERY_SECONDS": WORKER_QUERY_SECONDS,
        "PREFECT_DEFAULT_RESULT_STORAGE_BLOCK": WORKER_DEFAULT_RESULT_STORAGE_BLOCK,
    }
    labels: dict[str, str] = {
        "infrahub.app/version": infrahub_version,
    }


class InfrahubWorkerAsyncTemplateVariables(BaseVariables):
    pass


class InfrahubWorkerAsyncResult(BaseWorkerResult):
    """Result returned by the InfrahubWorker."""


class InfrahubWorkerAsync(BaseWorker):
    type: str = "infrahubasync"
    job_configuration = InfrahubWorkerAsyncConfiguration
    job_configuration_variables = InfrahubWorkerAsyncTemplateVariables
    _documentation_url = "https://example.com/docs"
    _logo_url = "https://example.com/logo"
    _description = "Infrahub worker designed to run the flow in the main async loop."
    service: InfrahubServices  # keep a reference to `service` so we can inject it within flows parameters.
    component_type = ComponentType.GIT_AGENT

    async def setup(
        self,
        client: InfrahubClient | None = None,
        metric_port: int | None = None,
        **kwargs: dict[str, Any],
    ) -> None:
        logging.getLogger("websockets").setLevel(logging.ERROR)
        logging.getLogger("httpx").setLevel(logging.ERROR)
        logging.getLogger("httpcore").setLevel(logging.ERROR)
        logging.getLogger("neo4j").setLevel(logging.ERROR)
        logging.getLogger("aio_pika").setLevel(logging.ERROR)
        logging.getLogger("aiormq").setLevel(logging.ERROR)
        logging.getLogger("git").setLevel(logging.ERROR)
        # Prevent git from interactively prompting the user for passwords if the credentials provided
        # by the credential helper is failing.
        os.environ["GIT_TERMINAL_PROMPT"] = "0"

        if not config.SETTINGS.settings:
            config_file = os.environ.get("INFRAHUB_CONFIG", "infrahub.toml")
            config.load_and_exit(config_file_name=config_file)

        self._init_logger()

        # Initialize trace
        if config.SETTINGS.trace.enable:
            configure_trace(
                service="infrahub-task-worker",
                version=infrahub_version,
                exporter_type=config.SETTINGS.trace.exporter_type,
                exporter_endpoint=config.SETTINGS.trace.exporter_endpoint,
                exporter_protocol=config.SETTINGS.trace.exporter_protocol,
            )

        # Start metric endpoint
        if metric_port is None or metric_port != 0:
            metric_port = metric_port or int(os.environ.get("INFRAHUB_METRICS_PORT", "8000"))
            self._logger.info(f"Starting metric endpoint on port {metric_port}")
            start_http_server(metric_port)

        await super().setup(**kwargs)

        self._exit_stack.enter_context(
            prefect_settings.temporary_settings(
                updates={  # type: ignore[arg-type]
                    prefect_settings.PREFECT_WORKER_QUERY_SECONDS: config.SETTINGS.workflow.worker_polling_interval,
                    prefect_settings.PREFECT_RESULTS_PERSIST_BY_DEFAULT: True,
                    prefect_settings.PREFECT_DEFAULT_RESULT_STORAGE_BLOCK: WORKER_DEFAULT_RESULT_STORAGE_BLOCK,
                }
            )
        )

        set_component_type(component_type=self.component_type)
        await self.set_git_global_config()
        await self._init_services(client=client)

        if not registry.schema_has_been_initialized():
            initialize_lock(service=self.service)

            async with self.service.database.start_session() as db:
                await initialization(db=db)

            await self.service.component.refresh_schema_hash()

        async with self.service.database.start_session() as dbs:
            await validate_graph_version(db=dbs)

        initialize_repositories_directory()
        build_component_registry()
        await self.service.scheduler.start_schedule()
        self._logger.info("Worker initialization completed .. ")

    async def run(
        self,
        flow_run: FlowRun,
        configuration: BaseJobConfiguration,
        task_status: TaskStatus[int] | None = None,
    ) -> BaseWorkerResult:
        flow_run_logger = self.get_flow_run_logger(flow_run)

        entrypoint: str = configuration._related_objects["deployment"].entrypoint

        file_path, flow_name = entrypoint.split(":")
        module_path = file_path.removeprefix("backend/").removesuffix(".py").replace("/", ".")
        flow_func = load_flow_function(module_path=module_path, flow_name=flow_name)
        inject_service_parameter(func=flow_func, parameters=flow_run.parameters, service=self.service)
        flow_run_logger.debug("Validating parameters")
        params = flow_func.validate_parameters(parameters=flow_run.parameters)

        if task_status:
            task_status.started(True)

        async with AsyncClientContext(httpx_settings={"verify": get_http().verify_tls()}) as ctx:
            ctx._httpx_settings = None  # Hack to make all child task/flow runs use the same client
            await run_flow_async(flow=flow_func, flow_run=flow_run, parameters=params, return_type="state")

        return InfrahubWorkerAsyncResult(status_code=0, identifier=str(flow_run.id))

    def _init_logger(self) -> None:
        """Initialize loggers to use the API handle provided by Prefect."""
        api_handler = APILogHandler()

        for logger_name in config.SETTINGS.workflow.extra_loggers + DEFAULT_TASK_LOGGERS:
            logger = logging.getLogger(logger_name)
            logger.setLevel(config.SETTINGS.workflow.extra_log_level.value)
            logger.addHandler(api_handler)

    async def _init_infrahub_client(self, client: InfrahubClient | None = None) -> InfrahubClient:
        if not client:
            self._logger.debug(f"Using Infrahub API at {config.SETTINGS.main.internal_address}")
            try:
                client = InfrahubClient(
                    config=Config(
                        address=config.SETTINGS.main.infrahub_address, retry_on_failure=True, log=self._logger
                    )
                )
            except InitializationError as err:
                self._logger.error(
                    "Infrahub client initialization failed due to missing configuration for internal_address."
                )
                raise typer.Exit(1) from err

        try:
            await client.branch.all()
        except SdkError as err:
            self._logger.error(f"Error in communication with Infrahub: {err.message}")
            raise typer.Exit(1) from err

        return client

    async def _init_services(self, client: InfrahubClient | None) -> None:
        client = await self._init_infrahub_client(client=client)

        service = await InfrahubServices.new(
            cache=await get_cache(),
            client=client,
            database=await get_database(),
            message_bus=await get_message_bus(),
            workflow=get_workflow(),
            component=await get_component(),
            component_type=self.component_type,
        )

        self.service = service

    async def set_git_global_config(self) -> None:
        global_config_file = config.SETTINGS.git.global_config_file
        if not os.getenv("GIT_CONFIG_GLOBAL") and global_config_file:
            config_dir = Path(global_config_file).parent
            with contextlib.suppress(FileExistsError):
                config_dir.mkdir(exist_ok=True, parents=True)
            os.environ["GIT_CONFIG_GLOBAL"] = global_config_file
            self._logger.info(f"Set git config file to {global_config_file}")

        await self._run_git_config_global(config.SETTINGS.git.user_name, setting_name="user.name")
        await self._run_git_config_global(config.SETTINGS.git.user_email, setting_name="user.email")
        await self._run_git_config_global("*", "--replace-all", setting_name="safe.directory")
        await self._run_git_config_global("true", setting_name="credential.usehttppath")
        await self._run_git_config_global(
            f"/usr/bin/env {config.SETTINGS.dev.git_credential_helper}", setting_name="credential.helper"
        )

    async def _run_git_config_global(self, *args: str, setting_name: str) -> None:
        proc = await asyncio.create_subprocess_exec(
            "git",
            "config",
            "--global",
            setting_name,
            *args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        _, stderr = await proc.communicate()
        if proc.returncode != 0:
            error_msg = stderr.decode("utf-8", errors="ignore").strip() or "unknown error"
            self._logger.error(f"Failed to set git {setting_name}: %s", error_msg)
        else:
            self._logger.info(f"Git {setting_name} set")



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING

from infrahub.core import registry
from infrahub.exceptions import ValidationError

from .model import SchemaViolation

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.path import DataPath, GroupedDataPaths
    from infrahub.database import InfrahubDatabase

    from .interface import ConstraintCheckerInterface
    from .model import SchemaConstraintValidatorRequest


class AggregatedConstraintChecker:
    def __init__(
        self, constraints: list[ConstraintCheckerInterface], db: InfrahubDatabase, branch: Branch | None = None
    ) -> None:
        self.constraints = constraints
        self.db = db
        self.branch = branch

    async def run_constraints(self, request: SchemaConstraintValidatorRequest) -> list[SchemaViolation]:
        grouped_data_paths_by_constraint_name: dict[str, list[GroupedDataPaths]] = {}
        for constraint in self.constraints:
            if constraint.supports(request):
                grouped_data_paths_by_constraint_name[constraint.name] = await constraint.check(request)

        ids: list[str] = []
        for grouped_path in chain(*grouped_data_paths_by_constraint_name.values()):
            ids.extend([path.node_id for path in grouped_path.get_all_data_paths()])
        # Try to query the nodes with their display label
        # it's possible that it might not work if the obj is not valid with the schema
        fields = {"display_label": None, request.schema_path.field_name: None}
        try:
            nodes = await registry.manager.get_many(db=self.db, ids=ids, branch=self.branch, fields=fields)
        except ValidationError:
            nodes = {}

        violations = []
        for constraint_name, grouped_paths in grouped_data_paths_by_constraint_name.items():
            for path in chain(*[gp.get_all_data_paths() for gp in grouped_paths]):
                node = nodes.get(path.node_id)
                node_display_label = None
                display_label = None
                if node:
                    node_display_label = await node.get_display_label(db=self.db)
                if node_display_label:
                    if request.node_schema.display_labels and node:
                        display_label = f"Node {node_display_label} ({node.get_kind()}: {path.node_id})"
                    else:
                        display_label = f"Node {node_display_label}"
                if not display_label:
                    display_label = f"Node ({path.kind}: {path.node_id})"

                violation = SchemaViolation(
                    node_id=path.node_id,
                    node_kind=path.kind,
                    display_label=node_display_label or display_label,
                    full_display_label=display_label,
                )
                violation.message = await self.render_error_request(
                    violation=violation, constraint_name=constraint_name, data_path=path
                )
                violations.append(violation)
        return violations

    async def render_error_request(
        self,
        violation: SchemaViolation,
        constraint_name: str,
        data_path: DataPath,
    ) -> str:
        constraint_name_str = constraint_name
        if constraint_name.count(".") == 2:
            constraint_level, constraint_name_str, _ = constraint_name.split(".", maxsplit=2)
            error_str = f"{constraint_level.title()}-level '{constraint_name_str}'"
        else:
            error_str = f"'{constraint_name_str}'"
        error_str += f" constraint violation on schema '{violation.node_kind}'."
        if violation.display_label.startswith("Node"):
            error_str += f" {violation.display_label}"
        else:
            error_str += f" Node ({violation.display_label})"
        error_str += " is not compliant."
        error_detail_str_list = []
        if data_path.field_name:
            if data_path.value:
                error_detail_str = data_path.field_name
                if data_path.property_name:
                    error_detail_str += f".{data_path.property_name}"
                error_detail_str += f"={data_path.value!r}"
                error_detail_str_list.append(error_detail_str)
            if data_path.peer_id:
                error_detail_str = f"{data_path.field_name}.id={data_path.peer_id}"
                error_detail_str_list.append(error_detail_str)
            if error_detail_str_list:
                error_str += " The error relates to field "
                error_str += ",".join(error_detail_str_list)
                error_str += "."
        return error_str



from infrahub.branch.status_checker import BranchStatusChecker

ALLOWED_MUTATIONS_ON_NEED_REBASE_BRANCH = ["BranchRebase", "BranchDelete", "BranchCreate", "ProposedChangeCreate"]
ALLOWED_MUTATIONS_ON_MERGED_BRANCH = ["BranchDelete"]


def raise_on_mutation_for_branch_status(next, root, info, **kwargs):  # type: ignore  # noqa
    if info.operation.operation.value == "mutation":
        mutation_name = info.operation.selection_set.selections[0].name.value
        brach_status_checker = BranchStatusChecker()
        if mutation_name not in ALLOWED_MUTATIONS_ON_NEED_REBASE_BRANCH:
            brach_status_checker.check_needs_rebase_status(branch=info.context.branch)
        if mutation_name not in ALLOWED_MUTATIONS_ON_MERGED_BRANCH:
            brach_status_checker.check_merge_status(branch=info.context.branch)

    return next(root, info, **kwargs)



from infrahub_sdk.protocols import CoreValidator

from infrahub.context import InfrahubContext
from infrahub.events.models import EventMeta
from infrahub.events.validator_action import ValidatorFailedEvent, ValidatorPassedEvent, ValidatorStartedEvent
from infrahub.services.adapters.event import InfrahubEventService


async def send_failed_validator(
    event_service: InfrahubEventService, validator: CoreValidator, proposed_change_id: str, context: InfrahubContext
) -> None:
    event = ValidatorFailedEvent(
        node_id=validator.id,
        kind=validator.get_kind(),
        proposed_change_id=proposed_change_id,
        meta=EventMeta.from_context(context=context),
    )
    await event_service.send(event=event)


async def send_passed_validator(
    event_service: InfrahubEventService, validator: CoreValidator, proposed_change_id: str, context: InfrahubContext
) -> None:
    event = ValidatorPassedEvent(
        node_id=validator.id,
        kind=validator.get_kind(),
        proposed_change_id=proposed_change_id,
        meta=EventMeta.from_context(context=context),
    )
    await event_service.send(event=event)


async def send_start_validator(
    event_service: InfrahubEventService, validator: CoreValidator, proposed_change_id: str, context: InfrahubContext
) -> None:
    event = ValidatorStartedEvent(
        node_id=validator.id,
        kind=validator.get_kind(),
        proposed_change_id=proposed_change_id,
        meta=EventMeta.from_context(context=context),
    )
    await event_service.send(event=event)



from abc import ABC, abstractmethod

from fast_depends import Depends, inject

from infrahub.auth import AccountSession
from infrahub.core.node import Node
from infrahub.database import InfrahubDatabase


class ProposedChangeChecker(ABC):
    # We can't use CoreProposedChange type instead of Node as fast_depends enforces pydantic runtime type checks.
    @abstractmethod
    async def verify_proposed_change_is_mergeable(
        self, proposed_change: Node, db: InfrahubDatabase, account_session: AccountSession
    ) -> None:
        """
        Raise an error if proposed change cannot be merged.
        """

        raise NotImplementedError()


class ProposedChangeCheckerCommunity(ProposedChangeChecker):
    async def verify_proposed_change_is_mergeable(
        self, proposed_change: Node, db: InfrahubDatabase, account_session: AccountSession
    ) -> None:
        pass


def get_proposed_change_checker() -> ProposedChangeChecker:
    return ProposedChangeCheckerCommunity()


# We can't use CoreProposedChange type instead of Node as fast_depends enforces pydantic runtime type checks.
@inject
async def verify_proposed_change_is_mergeable(
    proposed_change: Node,
    db: InfrahubDatabase,
    account_session: AccountSession,
    pc_checker: ProposedChangeChecker = Depends(get_proposed_change_checker),  # noqa: B008
) -> None:
    # type ignore due to fast-depends enforcing pydantic checks
    await pc_checker.verify_proposed_change_is_mergeable(
        proposed_change=proposed_change, db=db, account_session=account_session
    )



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, Optional, Self, Union, cast

from pydantic import Field, field_validator

from infrahub.core.branch.enums import BranchStatus
from infrahub.core.branch.filters import BranchListFilters
from infrahub.core.constants import GLOBAL_BRANCH_NAME, SYSTEM_USER_ID
from infrahub.core.graph import GRAPH_VERSION
from infrahub.core.models import SchemaBranchHash  # noqa: TC001
from infrahub.core.node.standard import StandardNode, StandardNodeOrdering
from infrahub.core.query import Query, QueryType
from infrahub.core.query.branch import (
    BranchNodeGetListQuery,
    DeleteBranchRelationshipsQuery,
    RebaseBranchQuery,
)
from infrahub.core.registry import registry
from infrahub.core.timestamp import Timestamp
from infrahub.exceptions import BranchNotFoundError, InitializationError, ValidationError

if TYPE_CHECKING:
    from neo4j.graph import Node as Neo4jNode

    from infrahub.database import InfrahubDatabase


class Branch(StandardNode):
    name: str = Field(
        max_length=250, min_length=3, description="Name of the branch (git ref standard)", validate_default=True
    )
    status: BranchStatus = BranchStatus.OPEN
    description: str = ""
    origin_branch: str = "main"
    branched_from: Optional[str] = Field(default=None, validate_default=True)
    hierarchy_level: int = 2
    is_default: bool = False
    is_global: bool = False
    is_protected: bool = False
    sync_with_git: bool = Field(
        default=True,
        description="Indicate if the branch should be extended to Git and if Infrahub should merge the branch in Git as part of a proposed change",
    )
    is_isolated: bool = True
    schema_changed_at: Optional[str] = None
    schema_hash: Optional[SchemaBranchHash] = None
    graph_version: int | None = None

    _exclude_attrs: list[str] = ["id", "uuid", "owner"]

    @field_validator("name", mode="before")
    @classmethod
    def validate_branch_name(cls, value: str) -> str:
        checks = [
            (r".*/\.", "/."),
            (r"\.\.", ".."),
            (r"^/", "starts with /"),
            (r"//", "//"),
            (r"@{", "@{"),
            (r"\\", "backslash (\\)"),
            (r"[\000-\037\177 ~^:?*[]", "disallowed ASCII characters/patterns"),
            (r"\.lock$", "ends with .lock"),
            (r"/$", "ends with /"),
            (r"\.$", "ends with ."),
        ]

        offending_patterns = []

        for pattern, description in checks:
            if re.search(pattern, value):
                offending_patterns.append(description)

        if value == GLOBAL_BRANCH_NAME:
            return value  # this is the only allowed exception

        if offending_patterns:
            error_text = ", ".join(offending_patterns)
            raise ValidationError(f"Branch name contains invalid patterns or characters: {error_text}")

        return value

    @field_validator("branched_from", mode="before")
    @classmethod
    def set_branched_from(cls, value: str) -> str:
        return Timestamp(value).to_string()

    def get_branched_from(self) -> str:
        if not self.branched_from:
            raise RuntimeError(f"branched_from not set for branch {self.name}")
        return self.branched_from

    def get_created_at(self) -> str:
        if not self.created_at:
            raise RuntimeError(f"created_at not set for branch {self.name}")
        return self.created_at

    @property
    def is_terminal(self) -> bool:
        return self.status in (BranchStatus.MERGED, BranchStatus.DELETING)

    @property
    def active_schema_hash(self) -> SchemaBranchHash:
        if self.schema_hash:
            return self.schema_hash

        raise InitializationError("The schema_hash has not been loaded for this branch")

    @property
    def has_schema_changes(self) -> bool:
        if not self.schema_hash:
            return False

        origin_branch = self.get_origin_branch()
        if not origin_branch or not origin_branch.schema_hash:
            return False

        if self.schema_hash.main != origin_branch.schema_hash.main:
            return True

        return False

    def update_schema_hash(self, at: Timestamp | str | None = None) -> bool:
        latest_schema = registry.schema.get_schema_branch(name=self.name)
        new_hash = latest_schema.get_hash_full()
        if self.schema_hash and new_hash.main == self.schema_hash.main:
            return False

        self.schema_changed_at = Timestamp(at).to_string()
        self.schema_hash = new_hash
        return True

    @classmethod
    async def get_by_name(cls, name: str, db: InfrahubDatabase, ignore_deleting: bool = True) -> Branch:
        query = """
        MATCH (n:Branch)
        WHERE n.name = $name
        AND NOT n.status IN $ignore_statuses
        RETURN n
        """

        params: dict[str, Any] = {"name": name}
        params["ignore_statuses"] = []
        if ignore_deleting:
            params["ignore_statuses"].append(BranchStatus.DELETING.value)

        results = await db.execute_query(query=query, params=params, name="branch_get_by_name", type=QueryType.READ)

        if len(results) == 0:
            raise BranchNotFoundError(identifier=name)

        return cls.from_db(results[0].values()[0])

    @classmethod
    async def get_list(
        cls,
        db: InfrahubDatabase,
        limit: int = 1000,
        ids: list[str] | None = None,
        name: str | None = None,
        node_ordering: StandardNodeOrdering | None = None,
        **kwargs: dict[str, Any],
    ) -> list[Self]:
        # Extract branch-specific params from kwargs, a future refactoring could update the parent signature for
        # standard nodes instead. Should be considered when additional StandardNode subclasses are introduced
        branch_filters: BranchListFilters | None = kwargs.pop("branch_filters", None)  # type: ignore[assignment]
        exclude_global: bool = kwargs.pop("exclude_global", False)  # type: ignore[assignment]

        if branch_filters is None:
            branch_filters = BranchListFilters(name=name, ids=ids)
        else:
            branch_filters.ids = branch_filters.ids or ids
            branch_filters.name = branch_filters.name or name

        node_ordering = node_ordering or StandardNodeOrdering()
        query: Query = await BranchNodeGetListQuery.init(
            db=db,
            node_class=cls,
            branch_filters=branch_filters,
            exclude_global=exclude_global,
            limit=limit,
            node_ordering=node_ordering,
        )
        await query.execute(db=db)

        return [cls.from_db(node=cast("Neo4jNode", result.get("n"))) for result in query.get_results()]

    @classmethod
    async def get_list_count(
        cls,
        db: InfrahubDatabase,
        limit: int = 1000,
        ids: list[str] | None = None,
        name: str | None = None,
        partial_match: bool = False,
        branch_filters: BranchListFilters | None = None,
        node_ordering: StandardNodeOrdering | None = None,
        **_kwargs: Any,
    ) -> int:
        if branch_filters is None:
            branch_filters = BranchListFilters(name=name, ids=ids, partial_match=partial_match)
        else:
            branch_filters.ids = branch_filters.ids or ids
            branch_filters.name = branch_filters.name or name
            branch_filters.partial_match = branch_filters.partial_match or partial_match

        node_ordering = node_ordering or StandardNodeOrdering()
        query: Query = await BranchNodeGetListQuery.init(
            db=db,
            node_class=cls,
            branch_filters=branch_filters,
            limit=limit,
            exclude_global=True,
            node_ordering=node_ordering,
        )
        return await query.count(db=db)

    @classmethod
    def isinstance(cls, obj: Any) -> bool:
        return isinstance(obj, cls)

    def get_origin_branch(self) -> Optional[Branch]:
        """Return the branch Object of the origin_branch."""
        if not self.origin_branch or self.origin_branch == self.name:
            return None

        return registry.get_branch_from_registry(branch=self.origin_branch)

    def get_branches_in_scope(self) -> list[str]:
        """Return the list of all the branches that are constituing this branch.

        For now, either a branch is the default branch or it must inherit from it so we can only have 2 values at best
        But the idea is that it will change at some point in a future version.
        """
        default_branch = registry.default_branch
        if self.name == default_branch:
            return [self.name]

        return [default_branch, self.name]

    def get_branches_and_times_to_query(self, at: Optional[Timestamp] = None) -> dict[frozenset, str]:
        """Return all the names of the branches that are constituing this branch with the associated times excluding the global branch"""

        at = Timestamp(at)

        if self.is_default:
            return {frozenset([self.name]): at.to_string()}

        time_default_branch = at

        # If the branch is isolated, and if the time requested is after the creation of the branch
        if self.is_isolated and at > Timestamp(self.branched_from):
            time_default_branch = Timestamp(self.branched_from)

        return {
            frozenset([self.origin_branch]): time_default_branch.to_string(),
            frozenset([self.name]): at.to_string(),
        }

    def get_branches_and_times_to_query_global(
        self,
        at: Optional[Timestamp] = None,
        is_isolated: bool = True,
    ) -> dict[frozenset, str]:
        """Return all the names of the branches that are constituting this branch with the associated times."""

        at = Timestamp(at)

        if self.is_default:
            return {frozenset((GLOBAL_BRANCH_NAME, self.name)): at.to_string()}

        time_default_branch = at

        # If the branch is isolated, and if the time requested is after the creation of the branch
        if self.is_isolated and is_isolated and at > Timestamp(self.branched_from):
            time_default_branch = Timestamp(self.branched_from)

        return {
            frozenset((GLOBAL_BRANCH_NAME, self.origin_branch)): time_default_branch.to_string(),
            frozenset((GLOBAL_BRANCH_NAME, self.name)): at.to_string(),
        }

    def get_branches_and_times_for_range(
        self, start_time: Timestamp, end_time: Timestamp
    ) -> tuple[dict[str, str], dict[str, str]]:
        """Return the names of the branches that are constituing this branch with the start and end times."""

        start = {}
        end = {}

        time_branched_from = Timestamp(self.branched_from)
        time_created_at = Timestamp(self.created_at)

        # Ensure start time is not older than the creation of the branch (time_created_at)
        time_query_start = start_time
        if start_time < time_created_at:
            time_query_start = time_created_at

        start[self.name] = time_query_start.to_string()

        # START
        if not self.is_default and time_query_start <= time_branched_from:
            start[self.origin_branch] = time_branched_from.to_string()
        elif not self.is_default and time_query_start > time_branched_from:
            start[self.origin_branch] = time_query_start.to_string()

        # END
        end[self.name] = end_time.to_string()
        if not self.is_default:
            end[self.origin_branch] = end_time.to_string()

        return start, end

    async def create(self, db: InfrahubDatabase, user_id: str = SYSTEM_USER_ID) -> bool:
        self.graph_version = GRAPH_VERSION
        return await super().create(db=db, user_id=user_id)

    async def delete(self, db: InfrahubDatabase) -> None:
        if self.is_default:
            raise ValidationError(f"Unable to delete {self.name} it is the default branch.")
        if self.is_global:
            raise ValidationError(f"Unable to delete {self.name} this is an internal branch.")

        self.status = BranchStatus.DELETING
        await self.save(db=db)

        query = await DeleteBranchRelationshipsQuery.init(db=db, branch_name=self.name)
        await query.execute(db=db)
        await super().delete(db=db)

    def get_query_filter_relationships(
        self, rel_labels: list, at: Optional[Timestamp] = None, include_outside_parentheses: bool = False
    ) -> tuple[list, dict]:
        """
        Generate a CYPHER Query filter based on a list of relationships to query a part of the graph at a specific time and on a specific branch.
        """

        filters = []
        params: dict[str, Any] = {}

        if not isinstance(rel_labels, list):
            raise TypeError(f"rel_labels must be a list, not a {type(rel_labels)}")

        at = Timestamp(at)
        branches_times = self.get_branches_and_times_to_query_global(at=at)

        for idx, (branch_name, time_to_query) in enumerate(branches_times.items()):
            params[f"branch{idx}"] = list(branch_name)
            params[f"time{idx}"] = time_to_query

        for rel in rel_labels:
            filters_per_rel = []
            for idx in range(len(branches_times)):
                filters_per_rel.append(
                    f"({rel}.branch IN $branch{idx} AND {rel}.from <= $time{idx} AND {rel}.to IS NULL)"
                )
                filters_per_rel.append(
                    f"({rel}.branch IN $branch{idx} AND {rel}.from <= $time{idx} AND {rel}.to > $time{idx})"
                )

            if not include_outside_parentheses:
                filters.append("\n OR ".join(filters_per_rel))

            filters.append("(" + "\n OR ".join(filters_per_rel) + ")")

        return filters, params

    def get_query_filter_path(
        self,
        at: Optional[Union[Timestamp, str]] = None,
        is_isolated: bool = True,
        branch_agnostic: bool = False,
        variable_name: str = "r",
        params_prefix: str = "",
    ) -> tuple[str, dict]:
        """
        Generate a CYPHER Query filter based on a path to query a part of the graph at a specific time and on a specific branch.

        Examples:
            >>> rels_filter, rels_params = self.branch.get_query_filter_path(at=self.at)
            >>> self.params.update(rels_params)
            >>> query += "\n WHERE all(r IN relationships(p) WHERE %s)" % rels_filter

            There is a currently an assumption that the relationship in the path will be named 'r'
        """
        pp = params_prefix
        params: dict[str, Any] = {}
        at = Timestamp(at)
        at_str = at.to_string()
        if branch_agnostic:
            filter_str = (
                f"{variable_name}.from < ${pp}time1 AND ({variable_name}.to IS NULL or {variable_name}.to > ${pp}time1)"
            )
            params[f"{pp}time1"] = at_str
            return filter_str, params

        branches_times = self.get_branches_and_times_to_query_global(at=at, is_isolated=is_isolated)

        for idx, (branch_name, time_to_query) in enumerate(branches_times.items()):
            params[f"{pp}branch{idx}"] = list(branch_name)
            params[f"{pp}time{idx}"] = time_to_query

        filters = []
        for idx in range(len(branches_times)):
            filters.append(
                f"({variable_name}.branch IN ${pp}branch{idx} "
                f"AND {variable_name}.from <= ${pp}time{idx} AND {variable_name}.to IS NULL)"
            )
            filters.append(
                f"({variable_name}.branch IN ${pp}branch{idx} "
                f"AND {variable_name}.from <= ${pp}time{idx} "
                f"AND {variable_name}.to > ${pp}time{idx})"
            )

        filter_str = "(" + "\n OR ".join(filters) + ")"

        return filter_str, params

    async def rebase(
        self, db: InfrahubDatabase, at: str | Timestamp | None = None, user_id: str = SYSTEM_USER_ID
    ) -> None:
        """Rebase the current Branch with its origin branch"""

        at = Timestamp(at)

        await self.rebase_graph(db=db, at=at)

        self.branched_from = at.to_string()
        self.status = BranchStatus.OPEN
        await self.save(db=db, user_id=user_id)

        # Update the branch in the registry after the rebase
        registry.branch[self.name] = self

    async def rebase_graph(self, db: InfrahubDatabase, at: Timestamp) -> None:
        """Rebase all relationships on this branch to a new point in time.

        This method updates the graph to reflect the state of the branch as if it had been created
        at the specified timestamp. Relationships are processed as follows:

        - Relationships with no `to` timestamp and `from` <= at: Updated to start from `at`
        - Relationships with `to` < at: Deleted (ended before rebase point)
        - Relationships with `to` >= at: Updated to start from `at`

        Orphaned nodes (nodes with no remaining relationships) are also cleaned up.
        """
        query = await RebaseBranchQuery.init(db=db, branch=self, at=at)
        await query.execute(db=db)


registry.branch_object = Branch



from __future__ import annotations

import ipaddress
from typing import TYPE_CHECKING

from graphene import Field, Int, ObjectType, String
from netaddr import IPSet

from infrahub.core.constants import InfrahubKind
from infrahub.core.ipam.resource_allocator import IPAMResourceAllocator
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import BuiltinIPPrefix
from infrahub.exceptions import NodeNotFoundError, ValidationError
from infrahub.pools.prefix import get_next_available_prefix

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.graphql.initialization import GraphqlContext


class IPAddressGetNextAvailable(ObjectType):
    address = String(required=True)

    @staticmethod
    async def resolve(
        root: dict,  # noqa: ARG004
        info: GraphQLResolveInfo,
        prefix_id: str,
        prefix_length: int | None = None,
    ) -> dict[str, str]:
        graphql_context: GraphqlContext = info.context

        prefix = await NodeManager.get_one(
            id=prefix_id, kind=BuiltinIPPrefix, db=graphql_context.db, branch=graphql_context.branch
        )

        if not prefix:
            raise NodeNotFoundError(
                branch_name=graphql_context.branch.name, node_type=InfrahubKind.IPPREFIX, identifier=prefix_id
            )

        ip_prefix = ipaddress.ip_network(prefix.prefix.value)  # type: ignore[attr-defined]
        prefix_length = prefix_length or ip_prefix.prefixlen

        if not ip_prefix.prefixlen <= prefix_length <= ip_prefix.max_prefixlen:
            raise ValidationError(input_value="Invalid prefix length for current selected prefix")

        namespace = await prefix.ip_namespace.get_peer(db=graphql_context.db)  # type: ignore[attr-defined]
        allocator = IPAMResourceAllocator(db=graphql_context.db, namespace=namespace, branch=graphql_context.branch)
        next_address = await allocator.get_next_address(
            ip_prefix=ip_prefix,
            is_pool=prefix.is_pool.value,  # type: ignore[attr-defined]
        )

        if not next_address:
            raise IndexError("No addresses available in prefix")

        return {"address": f"{next_address.ip}/{prefix_length}"}


class IPPrefixGetNextAvailable(ObjectType):
    prefix = String(required=True)

    @staticmethod
    async def resolve(
        root: dict,  # noqa: ARG004
        info: GraphQLResolveInfo,
        prefix_id: str,
        prefix_length: int | None = None,
    ) -> dict[str, str]:
        graphql_context: GraphqlContext = info.context

        prefix = await NodeManager.get_one(
            id=prefix_id, db=graphql_context.db, branch=graphql_context.branch, kind=BuiltinIPPrefix
        )

        if not prefix:
            raise NodeNotFoundError(
                branch_name=graphql_context.branch.name, node_type=InfrahubKind.IPPREFIX, identifier=prefix_id
            )

        ip_prefix = prefix.prefix.obj
        namespace = await prefix.ip_namespace.get_peer(db=graphql_context.db)
        allocator = IPAMResourceAllocator(db=graphql_context.db, namespace=namespace, branch=graphql_context.branch)

        # Build available pool by removing existing subnets from parent prefix
        subnets = await allocator.get_subnets(ip_prefix=ip_prefix)
        available_pool = IPSet([str(ip_prefix)])
        for subnet in subnets:
            available_pool.remove(str(subnet.prefix))

        if prefix_length is not None and not ip_prefix.prefixlen < prefix_length <= ip_prefix.max_prefixlen:
            raise ValidationError(input_value="Invalid prefix length for current selected prefix")

        try:
            next_prefix = get_next_available_prefix(
                pool=available_pool, prefix_length=prefix_length, prefix_ver=ip_prefix.version
            )
        except ValueError as exc:
            raise IndexError("No prefixes available in prefix") from exc

        return {"prefix": str(next_prefix)}


InfrahubIPAddressGetNextAvailable = Field(
    IPAddressGetNextAvailable,
    prefix_id=String(required=True),
    prefix_length=Int(required=False),
    resolver=IPAddressGetNextAvailable.resolve,
    required=True,
)


InfrahubIPPrefixGetNextAvailable = Field(
    IPPrefixGetNextAvailable,
    prefix_id=String(required=True),
    prefix_length=Int(required=False),
    resolver=IPPrefixGetNextAvailable.resolve,
    required=True,
)

# The following two query fields must be removed once we are sure that people are not using the old queries anymore. Those fields only exist to
# expose a deprecation message.

DeprecatedIPAddressGetNextAvailable = Field(
    IPAddressGetNextAvailable,
    prefix_id=String(required=True),
    prefix_length=Int(required=False),
    resolver=IPAddressGetNextAvailable.resolve,
    required=True,
    deprecation_reason="This query has been renamed to 'InfrahubIPAddressGetNextAvailable'. It will be removed in the next version of Infrahub.",
)


DeprecatedIPPrefixGetNextAvailable = Field(
    IPPrefixGetNextAvailable,
    prefix_id=String(required=True),
    prefix_length=Int(required=False),
    resolver=IPPrefixGetNextAvailable.resolve,
    required=True,
    deprecation_reason="This query has been renamed to 'InfrahubIPPrefixGetNextAvailable'. It will be removed in the next version of Infrahub.",
)



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core.constants import InfrahubKind

from .generated.genericnode_schema import GeneratedGenericSchema

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase


class GenericSchema(GeneratedGenericSchema):
    """A Generic can be either an Interface or a Union depending if there are some Attributes or Relationships defined."""

    @property
    def is_node_schema(self) -> bool:
        return False

    @property
    def is_generic_schema(self) -> bool:
        return True

    @property
    def is_profile_schema(self) -> bool:
        return False

    @property
    def is_template_schema(self) -> bool:
        return False

    @property
    def is_ip_prefix(self) -> bool:
        return self.kind == InfrahubKind.IPPREFIX

    @property
    def is_ip_address(self) -> bool:
        return self.kind == InfrahubKind.IPADDRESS

    def get_hierarchy_schema(self, db: InfrahubDatabase, branch: Branch | str | None = None) -> GenericSchema:  # noqa: ARG002
        if self.hierarchical:
            return self

        raise ValueError(f"hierarchical mode is not enabled on {self.kind}")

    def get_labels(self) -> list[str]:
        """Return the labels for this object"""
        return [self.kind]

    def _get_field_names_for_diff(self) -> list[str]:
        """Exclude used_by from the diff for generic nodes"""
        fields = super()._get_field_names_for_diff()
        return [field for field in fields if field != "used_by"]



from pydantic import Field

from infrahub.core.node.standard import StandardNode
from infrahub.core.query.standard_node import RootNodeCreateQuery, StandardNodeQuery


class Root(StandardNode):
    graph_version: int = Field(0, description="Internal Version of the graph.")
    default_branch: str = Field(..., description="The name of the default branch")

    _query: type[StandardNodeQuery] = RootNodeCreateQuery



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.constants import NULL_VALUE, RelationshipStatus
from infrahub.core.graph.schema import GraphAttributeValueIndexedNode, GraphAttributeValueNode
from infrahub.core.query import Query, QueryType
from infrahub.types import is_large_attribute_type

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class AttributeAddQuery(Query):
    name = "attribute_add"
    type = QueryType.WRITE

    def __init__(
        self,
        node_kinds: list[str],
        attribute_name: str,
        attribute_kind: str,
        branch_support: str,
        default_value: Any | None = None,
        uuids: list[str] | None = None,
        **kwargs: Any,
    ) -> None:
        self.node_kinds = node_kinds
        self.attribute_name = attribute_name
        self.attribute_kind = attribute_kind
        self.branch_support = branch_support
        self.default_value = default_value
        self.uuids = uuids
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        self.params["node_kinds"] = self.node_kinds
        self.params["attr_name"] = self.attribute_name
        self.params["branch_support"] = self.branch_support
        self.params["current_time"] = self.at.to_string()

        if self.default_value is not None:
            self.params["attr_value"] = self.default_value
        else:
            self.params["attr_value"] = NULL_VALUE

        self.params["user_id"] = self.user_id

        self.params["rel_props"] = {
            "branch": self.branch.name,
            "branch_level": self.branch.hierarchy_level,
            "status": RelationshipStatus.ACTIVE.value,
            "from": self.at.to_string(),
            "from_user_id": self.user_id,
        }

        self.params["is_protected_default"] = False

        # Set metadata for vertex properties on default/global branch
        self.params["set_metadata"] = self.branch.is_default or self.branch.is_global

        attr_value_label = GraphAttributeValueNode.get_default_label()
        if not is_large_attribute_type(self.attribute_kind):
            # should be indexed
            attr_value_label += f":{GraphAttributeValueIndexedNode.get_default_label()}"
            match_query = """
            MERGE (av:%(attr_value_label)s { value: $attr_value, is_default: true })
            LIMIT 1
            """ % {"attr_value_label": attr_value_label}
        else:
            # cannot be indexed
            match_query = """
            OPTIONAL MATCH (existing_av:%(attr_value_label)s { value: $attr_value, is_default: true })
            WHERE NOT existing_av:AttributeValueIndexed
            CALL (existing_av) {
                WITH existing_av
                WHERE existing_av IS NULL
                CREATE (:%(attr_value_label)s { value: $attr_value, is_default: true })
            }
            MATCH (av:%(attr_value_label)s { value: $attr_value, is_default: true })
            WHERE NOT av:AttributeValueIndexed
            LIMIT 1
            """ % {"attr_value_label": attr_value_label}

        node_kinds_str = "|".join(self.node_kinds)
        query = """
        %(match_query)s
        MERGE (is_protected_value:Boolean { value: $is_protected_default })
        WITH av, is_protected_value
        MATCH (n:%(node_kinds_str)s)
        CALL (n) {
            MATCH (:Root)<-[r:IS_PART_OF]-(n)
            WHERE %(branch_filter)s
            WITH n, r AS is_part_of_e
            OPTIONAL MATCH (n)-[r:HAS_ATTRIBUTE]-(:Attribute { name: $attr_name })
            WHERE %(branch_filter)s
            WITH is_part_of_e, r AS has_attr_e
            RETURN is_part_of_e, has_attr_e
            ORDER BY has_attr_e.branch_level DESC, has_attr_e.from ASC, is_part_of_e.branch_level DESC, is_part_of_e.from ASC
            LIMIT 1
        }
        WITH n, is_part_of_e, has_attr_e, av, is_protected_value
        WHERE is_part_of_e.status = "active" AND (has_attr_e IS NULL OR has_attr_e.status = "deleted")
        CREATE (a:Attribute { name: $attr_name, branch_support: $branch_support })
        CREATE (n)-[:HAS_ATTRIBUTE $rel_props ]->(a)
        CREATE (a)-[:HAS_VALUE $rel_props ]->(av)
        CREATE (a)-[:IS_PROTECTED $rel_props]->(is_protected_value)
        %(uuid_generation)s
        // Set metadata on Attribute and Node vertices if on default/global branch
        WITH a, n, has_attr_e
        CALL (a, n) {
            WITH a, n
            WHERE $set_metadata
            SET a.created_at = $current_time, a.created_by = $user_id, a.updated_at = $current_time, a.updated_by = $user_id
            SET n.updated_at = $current_time, n.updated_by = $user_id
        }
        FOREACH (i in CASE WHEN has_attr_e.status = "deleted" THEN [1] ELSE [] END |
            SET has_attr_e.to = $current_time, has_attr_e.to_user_id = $user_id
        )
        """ % {
            "match_query": match_query,
            "branch_filter": branch_filter,
            "node_kinds_str": node_kinds_str,
            "uuid_generation": db.render_uuid_generation(node_label="a", node_attr="uuid"),
        }

        self.add_to_query(query)
        self.return_labels = ["n.uuid", "a.uuid"]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.constants import GLOBAL_BRANCH_NAME
from infrahub.core.migrations.shared import GraphMigration, MigrationInput, MigrationResult
from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class FixGlobalBranchLevelQuery(Query):
    """Update edges on the global branch to have branch_level = 1."""

    name = "fix_global_branch_level"
    type: QueryType = QueryType.WRITE
    insert_return = False
    raise_error_if_empty = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["global_branch_name"] = GLOBAL_BRANCH_NAME

        query = """
        MATCH ()-[e {branch: $global_branch_name}]->()
        WHERE e.branch_level <> 1
        CALL (e) {
            SET e.branch_level = 1
        } IN TRANSACTIONS
        """
        self.add_to_query(query)


class Migration052(GraphMigration):
    """
    Fix edges on the global branch that have incorrect branch_level.

    Edges on the global branch ("-global-") should always have branch_level = 1.
    This migration corrects any edges that were incorrectly created with a different
    branch_level value (e.g., branch_level = 2 when created from a user branch).
    """

    name: str = "052_fix_global_branch_level"
    minimum_version: int = 51
    queries: Sequence[type[Query]] = [FixGlobalBranchLevelQuery]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        return await self.do_execute(migration_input=migration_input)



from __future__ import annotations

from typing import TYPE_CHECKING

from rich.progress import Progress

from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.initialization import get_root_node
from infrahub.core.migrations.graph.m044_backfill_hfid_display_label_in_db import DefaultBranchNodeCount, Migration044
from infrahub.core.migrations.shared import MigrationInput, MigrationResult, get_migration_console
from infrahub.exceptions import SchemaNotFoundError

from .load_schema_branch import get_or_load_schema_branch

if TYPE_CHECKING:
    from infrahub.core.schema import ProfileSchema, TemplateSchema


console = get_migration_console()


class Migration045(Migration044):
    """
    Backfill `human_friendly_id` and `display_label` attributes for profile and template nodes with schemas that define them.
    """

    name: str = "045_backfill_hfid_display_label_in_db_profile_template"
    minimum_version: int = 44
    update_batch_size: int = 1000

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        at = migration_input.at
        root_node = await get_root_node(db=db, initialize=False)
        default_branch_name = root_node.default_branch
        default_branch = await Branch.get_by_name(db=db, name=default_branch_name)

        main_schema_branch = await get_or_load_schema_branch(db=db, branch=default_branch)
        kinds_to_skip = self.kinds_to_skip + main_schema_branch.node_names

        total_nodes_query = await DefaultBranchNodeCount.init(db=db, kinds_to_skip=kinds_to_skip)
        await total_nodes_query.execute(db=db)
        total_nodes_count = total_nodes_query.get_num_nodes()

        base_node_schema = main_schema_branch.get("SchemaNode", duplicate=False)
        display_label_attribute_schema = base_node_schema.get_attribute("display_label")
        display_labels_attribute_schema = base_node_schema.get_attribute("display_labels")
        hfid_attribute_schema = base_node_schema.get_attribute("human_friendly_id")

        try:
            with Progress(console=console) as progress:
                update_task = progress.add_task(
                    f"Set display_label and human_friendly_id for {total_nodes_count} nodes on default branch",
                    total=total_nodes_count,
                )
                for node_schema_name in main_schema_branch.profile_names + main_schema_branch.template_names:
                    if node_schema_name in self.kinds_to_skip:
                        continue

                    node_schema: ProfileSchema | TemplateSchema
                    if node_schema_name in main_schema_branch.profile_names:
                        node_schema = main_schema_branch.get_profile(name=node_schema_name, duplicate=False)
                    else:
                        node_schema = main_schema_branch.get_template(name=node_schema_name, duplicate=False)

                    attribute_schema_map = {}
                    if node_schema.display_labels:
                        attribute_schema_map[display_labels_attribute_schema] = display_label_attribute_schema
                    if node_schema.human_friendly_id:
                        attribute_schema_map[hfid_attribute_schema] = hfid_attribute_schema
                    if not attribute_schema_map:
                        continue

                    await self._do_one_schema_all(
                        db=db,
                        branch=default_branch,
                        schema=node_schema,
                        schema_branch=main_schema_branch,
                        attribute_schema_map=attribute_schema_map,
                        at=at,
                        progress=progress,
                        update_task=update_task,
                    )

        except Exception as exc:
            return MigrationResult(errors=[str(exc)])
        return MigrationResult()

    async def execute_against_branch(self, migration_input: MigrationInput, branch: Branch) -> MigrationResult:
        db = migration_input.db
        at = migration_input.at
        default_branch = await Branch.get_by_name(db=db, name=registry.default_branch)
        main_schema_branch = await get_or_load_schema_branch(db=db, branch=default_branch)
        schema_branch = await get_or_load_schema_branch(db=db, branch=branch)

        base_node_schema = schema_branch.get("SchemaNode", duplicate=False)
        display_label_attribute_schema = base_node_schema.get_attribute("display_label")
        display_labels_attribute_schema = base_node_schema.get_attribute("display_labels")
        hfid_attribute_schema = base_node_schema.get_attribute("human_friendly_id")

        try:
            for node_schema_name in schema_branch.profile_names + schema_branch.template_names:
                if node_schema_name in self.kinds_to_skip:
                    continue

                node_schema: ProfileSchema | TemplateSchema
                default_node_schema: ProfileSchema | TemplateSchema | None
                if node_schema_name in schema_branch.profile_names:
                    node_schema = schema_branch.get_profile(name=node_schema_name, duplicate=False)
                    try:
                        default_node_schema = main_schema_branch.get_profile(name=node_schema_name, duplicate=False)
                    except SchemaNotFoundError:
                        default_node_schema = None
                else:
                    node_schema = schema_branch.get_template(name=node_schema_name, duplicate=False)
                    try:
                        default_node_schema = main_schema_branch.get_template(name=node_schema_name, duplicate=False)
                    except SchemaNotFoundError:
                        default_node_schema = None

                schemas_for_universal_update_map = {}
                schemas_for_targeted_update_map = {}
                if node_schema.display_labels:
                    if default_node_schema is None or default_node_schema.display_label != node_schema.display_label:
                        schemas_for_universal_update_map[display_labels_attribute_schema] = (
                            display_label_attribute_schema
                        )
                    else:
                        schemas_for_targeted_update_map[display_labels_attribute_schema] = (
                            display_label_attribute_schema
                        )

                if node_schema.human_friendly_id:
                    if (
                        default_node_schema is None
                        or default_node_schema.human_friendly_id != node_schema.human_friendly_id
                    ):
                        schemas_for_universal_update_map[hfid_attribute_schema] = hfid_attribute_schema
                    else:
                        schemas_for_targeted_update_map[hfid_attribute_schema] = hfid_attribute_schema

                if schemas_for_universal_update_map:
                    await self._do_one_schema_all(
                        db=db,
                        branch=branch,
                        schema=node_schema,
                        schema_branch=schema_branch,
                        attribute_schema_map=schemas_for_universal_update_map,
                        at=at,
                    )

                if not schemas_for_targeted_update_map:
                    continue

                for source_attribute_schema, destination_attribute_schema in schemas_for_targeted_update_map.items():
                    await self._do_one_schema_branch(
                        db=db,
                        branch=branch,
                        schema=node_schema,
                        schema_branch=schema_branch,
                        source_attribute_schema=source_attribute_schema,
                        destination_attribute_schema=destination_attribute_schema,
                        at=at,
                    )

        except Exception as exc:
            return MigrationResult(errors=[str(exc)])
        return MigrationResult()



from infrahub.core.branch import Branch
from infrahub.core.branch.enums import BranchStatus
from infrahub.exceptions import BranchAlreadyMergedError, BranchNeedsRebaseError


class BranchStatusChecker:
    def check_merge_status(self, branch: Branch) -> None:
        if branch.status == BranchStatus.MERGED:
            raise BranchAlreadyMergedError(
                identifier=branch.name,
                message=f"Branch '{branch.name}' has been merged and is read-only. No modifications are allowed.",
            )

    def check_needs_rebase_status(self, branch: Branch) -> None:
        if branch.status == BranchStatus.NEED_REBASE:
            raise BranchNeedsRebaseError(
                identifier=branch.name, message=f"Branch {branch.name} must be rebased before any updates can be made"
            )

    def check(self, branch: Branch) -> None:
        self.check_needs_rebase_status(branch)
        self.check_merge_status(branch)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from infrahub.core.branch import Branch
from infrahub.core.constants import NULL_VALUE, InfrahubKind, MetadataOptions
from infrahub.core.initialization import get_root_node
from infrahub.core.manager import NodeManager
from infrahub.core.migrations.query.update_attribute_values import UpdateAttributeValuesQuery
from infrahub.core.migrations.shared import (
    MigrationInput,
    MigrationRequiringRebase,
    MigrationResult,
    get_migration_console,
)

from .load_schema_branch import get_or_load_schema_branch

if TYPE_CHECKING:
    from infrahub.core.node import Node
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.database import InfrahubDatabase

console = get_migration_console()


class Migration063(MigrationRequiringRebase):
    """Nullify attribute values on template nodes where the source is a CoreNumberPool."""

    name: str = "063_template_number_pool_cleanup"
    minimum_version: int = 62

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        root_node = await get_root_node(db=db, initialize=False)
        default_branch_name = root_node.default_branch
        default_branch = await Branch.get_by_name(db=db, name=default_branch_name)
        schema_branch = await get_or_load_schema_branch(db=db, branch=default_branch)
        return await self._process_templates(db=db, branch=default_branch, schema_branch=schema_branch)

    async def execute_against_branch(self, migration_input: MigrationInput, branch: Branch) -> MigrationResult:
        db = migration_input.db
        schema_branch = await get_or_load_schema_branch(db=db, branch=branch)
        return await self._process_templates(db=db, branch=branch, schema_branch=schema_branch)

    async def _process_templates(
        self, db: InfrahubDatabase, branch: Branch, schema_branch: SchemaBranch
    ) -> MigrationResult:
        try:
            for template_kind in schema_branch.template_names:
                template_schema = schema_branch.get(name=template_kind, duplicate=False)

                templates = await NodeManager.query(
                    db=db,
                    schema=template_kind,
                    branch=branch,
                    include_metadata=MetadataOptions.LINKED_NODES,
                )

                # Collect node IDs to nullify, grouped by attribute name
                nullify_map: dict[str, set[str]] = defaultdict(set)

                for template in templates:
                    attr_names_to_nullify = await self._collect_pool_sourced_attrs(
                        db=db, template=template, template_schema=template_schema
                    )
                    for attr_name in attr_names_to_nullify:
                        nullify_map[attr_name].add(template.id)

                # Write updates via UpdateAttributeValuesQuery (bypasses ORM validation)
                for attr_name, node_ids in nullify_map.items():
                    attr_schema = template_schema.get_attribute(attr_name)
                    values_by_id = dict.fromkeys(node_ids, NULL_VALUE)
                    query = await UpdateAttributeValuesQuery.init(
                        db=db, branch=branch, attribute_schema=attr_schema, values_by_id_map=values_by_id
                    )
                    await query.execute(db=db)

        except Exception as exc:
            error_msg = str(exc) or f"{type(exc).__name__}: {repr(exc)}"
            return MigrationResult(errors=[error_msg])

        return MigrationResult()

    async def _collect_pool_sourced_attrs(
        self,
        db: InfrahubDatabase,
        template: Node,
        template_schema: MainSchemaTypes,
    ) -> list[str]:
        attrs_to_nullify: list[str] = []
        for attr_schema in template_schema.attributes:
            attribute = template.get_attribute(attr_schema.name)
            if not attribute or not getattr(attribute, "source_id", None):
                continue
            source = await attribute.get_source(db=db)
            if not source or source.get_kind() != InfrahubKind.NUMBERPOOL:
                continue
            if attribute.value is not None and attribute.value != NULL_VALUE:
                attrs_to_nullify.append(attr_schema.name)
        return attrs_to_nullify



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, Any

from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


class RelationshipDirection(Enum):
    OUTBOUND = "outbound"
    INBOUND = "inbound"
    BIDIRECTIONAL = "bidirectional"


@dataclass
class AttributeMetadata:
    uuid: str
    name: str
    is_deleted: bool
    created_at: Timestamp | None = None
    created_by: str | None = None
    updated_at: Timestamp | None = None
    updated_by: str | None = None

    def __hash__(self) -> int:
        return hash(
            (
                self.uuid,
                self.name,
                self.is_deleted,
                self.created_at,
                self.created_by,
                self.updated_at,
                self.updated_by,
            )
        )


@dataclass
class RelationshipMetadata:
    uuid: str
    identifier: str  # Relationship.name in DB (e.g., "testcar__testperson")
    peer_uuid: str
    direction: RelationshipDirection
    is_deleted: bool
    created_at: Timestamp | None = None
    created_by: str | None = None
    updated_at: Timestamp | None = None
    updated_by: str | None = None

    def __hash__(self) -> int:
        return hash(
            (
                self.uuid,
                self.identifier,
                self.peer_uuid,
                self.direction,
                self.is_deleted,
                self.created_at,
                self.created_by,
                self.updated_at,
                self.updated_by,
            )
        )


@dataclass
class NodeMetadata:
    uuid: str
    kind: str
    is_deleted: bool
    created_at: Timestamp | None = None
    created_by: str | None = None
    updated_at: Timestamp | None = None
    updated_by: str | None = None
    attributes: list[AttributeMetadata] = field(default_factory=list)
    relationships: list[RelationshipMetadata] = field(default_factory=list)

    def __hash__(self) -> int:
        return hash(
            (
                self.uuid,
                self.kind,
                self.is_deleted,
                self.created_at,
                self.created_by,
                self.updated_at,
                self.updated_by,
                tuple(self.attributes),
                tuple(self.relationships),
            )
        )


class NodeMetadataDefaultBranchQuery(Query):
    """Query to retrieve metadata for nodes and their attributes/relationships.

    This query only works on the default branch and reads metadata directly from
    vertex properties. It supports retrieving deleted nodes, attributes, and
    relationships.
    """

    name = "node_metadata"
    type = QueryType.READ
    insert_return = False

    def __init__(self, node_uuids: list[str], **kwargs: Any) -> None:
        self.node_uuids = node_uuids
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        if not self.branch.is_default:
            raise ValueError("NodeMetadataQuery only runs on the default branch")

        self.params["node_uuids"] = self.node_uuids
        self.params["branch"] = self.branch.name

        # Query nodes with their metadata and deletion status
        # Then query attributes and relationships
        query = """
// ------------------
// Part 1: Get node metadata with deletion status
// ------------------
UNWIND $node_uuids AS node_uuid
CALL (node_uuid) {
    MATCH (n:Node {uuid: node_uuid})-[r_ipo:IS_PART_OF]->(root:Root)
    WHERE r_ipo.branch = $branch
    RETURN n, r_ipo
    ORDER BY r_ipo.from DESC
    LIMIT 1
}
WITH n,
    CASE
        WHEN r_ipo.status = "deleted" THEN true
        WHEN r_ipo.to IS NOT NULL THEN true
        ELSE false
    END AS node_is_deleted

// ------------------
// Part 2: Get attribute details
// ------------------
OPTIONAL MATCH (n)-[:HAS_ATTRIBUTE {branch: $branch}]->(attr:Attribute)
WITH DISTINCT n, node_is_deleted, attr
CALL (n, attr) {
    OPTIONAL MATCH (n)-[r_attr:HAS_ATTRIBUTE]->(attr)
    WHERE r_attr.branch = $branch
    WITH r_attr, attr
    ORDER BY r_attr.from DESC
    LIMIT 1
    RETURN {
        uuid: attr.uuid,
        name: attr.name,
        is_deleted: CASE
            WHEN r_attr.status = "deleted" THEN true
            WHEN r_attr.to IS NOT NULL THEN true
            ELSE false
        END,
        created_at: attr.created_at,
        created_by: attr.created_by,
        updated_at: attr.updated_at,
        updated_by: attr.updated_by
    } AS attribute_details
}
WITH n, node_is_deleted, COALESCE(collect(attribute_details), []) AS attributes

// ------------------
// Part 3: Get relationship details
// ------------------
OPTIONAL MATCH (n)-[:IS_RELATED {branch: $branch}]-(rel:Relationship)-[:IS_RELATED {branch: $branch}]-(peer:Node)
WHERE n <> peer
WITH DISTINCT n, node_is_deleted, attributes, rel, peer
CALL (n, rel, peer) {
    OPTIONAL MATCH (n)-[r1:IS_RELATED]-(rel:Relationship)-[r2:IS_RELATED]-(peer:Node)
    WHERE r1.branch = $branch AND r2.branch = $branch
    WITH r1, r2
    ORDER BY r1.from DESC, r2.from DESC
    LIMIT 1
    RETURN {
        uuid: rel.uuid,
        identifier: rel.name,
        peer_uuid: peer.uuid,
        direction: CASE
            WHEN startNode(r1) = n AND startNode(r2) = rel THEN "outbound"
            WHEN startNode(r1) = rel AND startNode(r2) = peer THEN "inbound"
            ELSE "bidirectional"
        END,
        is_deleted: CASE
            WHEN (r1.status = "deleted" OR r1.to IS NOT NULL) AND (r2.status = "deleted" OR r2.to IS NOT NULL) THEN true
            ELSE false
        END,
        created_at: rel.created_at,
        created_by: rel.created_by,
        updated_at: rel.updated_at,
        updated_by: rel.updated_by
    } AS relationship_details
}
WITH n, node_is_deleted, attributes, COALESCE(collect(relationship_details), []) AS relationships

RETURN n.uuid AS node_uuid, n.kind AS node_kind, node_is_deleted,
       n.created_at AS node_created_at, n.created_by AS node_created_by,
       n.updated_at AS node_updated_at, n.updated_by AS node_updated_by,
       attributes, relationships
        """
        self.return_labels = [
            "node_uuid",
            "node_kind",
            "node_is_deleted",
            "node_created_at",
            "node_created_by",
            "node_updated_at",
            "node_updated_by",
            "attributes",
            "relationships",
        ]
        self.add_to_query(query)

    def get_metadatas(self) -> list[NodeMetadata]:
        """Process query results into NodeMetadata dataclasses."""
        from infrahub.core.timestamp import Timestamp

        nodes: list[NodeMetadata] = []

        for result in self.get_results():
            node_uuid = result.get_as_type("node_uuid", return_type=str)
            node_kind = result.get_as_type("node_kind", return_type=str)
            node_is_deleted = result.get_as_type("node_is_deleted", bool)

            node_created_at_str = result.get_as_str("node_created_at")
            node_created_at = Timestamp(node_created_at_str) if node_created_at_str else None
            node_created_by = result.get_as_str("node_created_by")
            node_updated_at_str = result.get_as_str("node_updated_at")
            node_updated_at = Timestamp(node_updated_at_str) if node_updated_at_str else None
            node_updated_by = result.get_as_str("node_updated_by")

            # Parse attributes
            attributes_data: list[dict[str, Any]] = result.get_as_type("attributes", list)
            attributes: list[AttributeMetadata] = []
            for attr_data in attributes_data:
                attr_created_at_str = attr_data.get("created_at")
                attr_created_at = Timestamp(attr_created_at_str) if attr_created_at_str else None
                attr_updated_at_str = attr_data.get("updated_at")
                attr_updated_at = Timestamp(attr_updated_at_str) if attr_updated_at_str else None

                attributes.append(
                    AttributeMetadata(
                        uuid=attr_data["uuid"],
                        name=attr_data["name"],
                        is_deleted=attr_data["is_deleted"],
                        created_at=attr_created_at,
                        created_by=attr_data.get("created_by"),
                        updated_at=attr_updated_at,
                        updated_by=attr_data.get("updated_by"),
                    )
                )

            # Parse relationships
            relationships_data: list[dict[str, Any]] = result.get_as_type("relationships", list)
            relationships: list[RelationshipMetadata] = []
            for rel_data in relationships_data:
                rel_created_at_str = rel_data.get("created_at")
                rel_created_at = Timestamp(rel_created_at_str) if rel_created_at_str else None
                rel_updated_at_str = rel_data.get("updated_at")
                rel_updated_at = Timestamp(rel_updated_at_str) if rel_updated_at_str else None

                direction_str = rel_data["direction"]
                direction = RelationshipDirection(direction_str)

                relationships.append(
                    RelationshipMetadata(
                        uuid=rel_data["uuid"],
                        identifier=rel_data["identifier"],
                        peer_uuid=rel_data["peer_uuid"],
                        direction=direction,
                        is_deleted=rel_data["is_deleted"],
                        created_at=rel_created_at,
                        created_by=rel_data.get("created_by"),
                        updated_at=rel_updated_at,
                        updated_by=rel_data.get("updated_by"),
                    )
                )

            nodes.append(
                NodeMetadata(
                    uuid=node_uuid,
                    kind=node_kind,
                    is_deleted=node_is_deleted,
                    created_at=node_created_at,
                    created_by=node_created_by,
                    updated_at=node_updated_at,
                    updated_by=node_updated_by,
                    attributes=attributes,
                    relationships=relationships,
                )
            )

        return nodes



import asyncio
from typing import TYPE_CHECKING, Any, AsyncGenerator

from graphene import Schema
from graphql import GraphQLResolveInfo, graphql

from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreGraphQLQuery
from infrahub.core.timestamp import Timestamp
from infrahub.graphql.resolvers.account_metadata import AccountMetadataResolver
from infrahub.graphql.resolvers.many_relationship import ManyRelationshipResolver
from infrahub.graphql.resolvers.single_relationship import SingleRelationshipResolver
from infrahub.log import get_logger

if TYPE_CHECKING:
    from infrahub.graphql.initialization import GraphqlContext

log = get_logger(name="infrahub.graphql")


async def resolver_graphql_query(
    parent: dict,  # noqa: ARG001
    info: GraphQLResolveInfo,
    name: str,
    graphql_schema: Schema,
    params: dict[str, Any] | None = None,
    interval: int = 10,
) -> AsyncGenerator[dict[str, Any], None]:
    graphql_context: GraphqlContext = info.context
    at = Timestamp()

    async with graphql_context.db.start_session(read_only=True) as db:
        # Find the GraphQLQuery and the GraphQL Schema
        graphql_query = await NodeManager.get_one_by_default_filter(
            db=db, id=name, kind=CoreGraphQLQuery, branch=graphql_context.branch, at=at
        )
        if not graphql_query:
            raise ValueError(f"Unable to find the {InfrahubKind.GRAPHQLQUERY} {name}")

    while True:
        async with graphql_context.db.start_session(read_only=True) as db:
            result = await graphql(
                schema=graphql_schema,
                source=graphql_query.query.value,
                context_value=graphql_context.__class__(
                    db=db,
                    branch=graphql_context.branch,
                    at=Timestamp(),
                    related_node_ids=set(),
                    types=graphql_context.types,
                    single_relationship_resolver=SingleRelationshipResolver(),
                    many_relationship_resolver=ManyRelationshipResolver(),
                    account_metadata_resolver=AccountMetadataResolver(),
                ),
                root_value=None,
                variable_values=params or {},
            )
            if result.data:
                yield result.data

        await asyncio.sleep(delay=float(interval))



# Generated by "invoke backend.generate", do not edit directly

from __future__ import annotations

from typing import TYPE_CHECKING

from .protocols_base import CoreNode, CoreNodeSync

if TYPE_CHECKING:
    from infrahub_sdk.node import RelatedNode, RelatedNodeSync, RelationshipManager, RelationshipManagerSync

    from .protocols_base import (
        URL,
        Boolean,
        BooleanOptional,
        DateTime,
        DateTimeOptional,
        Dropdown,
        Enum,
        HashedPassword,
        Integer,
        IntegerOptional,
        IPHost,
        IPNetwork,
        JSONAttribute,
        JSONAttributeOptional,
        ListAttributeOptional,
        String,
        StringOptional,
    )


# ---------------------------------------------
# ASYNC
# ---------------------------------------------


class BuiltinIPAddress(CoreNode):
    address: IPHost
    description: StringOptional
    ip_namespace: RelatedNode
    ip_prefix: RelatedNode


class BuiltinIPNamespace(CoreNode):
    name: String
    description: StringOptional
    ip_prefixes: RelationshipManager
    ip_addresses: RelationshipManager


class BuiltinIPPrefix(CoreNode):
    prefix: IPNetwork
    description: StringOptional
    member_type: Dropdown
    is_pool: Boolean
    is_top_level: BooleanOptional
    utilization: IntegerOptional
    netmask: StringOptional
    hostmask: StringOptional
    network_address: StringOptional
    broadcast_address: StringOptional
    ip_namespace: RelatedNode
    ip_addresses: RelationshipManager
    resource_pool: RelationshipManager
    parent: RelatedNode
    children: RelationshipManager


class CoreAction(CoreNode):
    name: String
    description: StringOptional
    triggers: RelationshipManager


class CoreArtifactTarget(CoreNode):
    artifacts: RelationshipManager


class CoreBasePermission(CoreNode):
    description: StringOptional
    identifier: StringOptional
    roles: RelationshipManager


class CoreCheck(CoreNode):
    name: StringOptional
    label: StringOptional
    origin: String
    kind: String
    message: StringOptional
    conclusion: Enum
    severity: Enum
    created_at: DateTimeOptional
    validator: RelatedNode


class CoreComment(CoreNode):
    text: String
    created_at: DateTimeOptional
    created_by: RelatedNode


class CoreCredential(CoreNode):
    name: String
    label: StringOptional
    description: StringOptional


class CoreFileObject(CoreNode):
    file_name: String
    checksum: String
    file_size: Integer
    file_type: String
    storage_id: String


class CoreGenericAccount(CoreNode):
    name: String
    password: HashedPassword
    label: StringOptional
    description: StringOptional
    account_type: Enum
    status: Dropdown
    tokens: RelationshipManager


class CoreGenericRepository(CoreNode):
    name: String
    description: StringOptional
    location: String
    internal_status: Dropdown
    operational_status: Dropdown
    sync_status: Dropdown
    credential: RelatedNode
    tags: RelationshipManager
    transformations: RelationshipManager
    queries: RelationshipManager
    checks: RelationshipManager
    generators: RelationshipManager
    groups_objects: RelationshipManager


class CoreGroup(CoreNode):
    name: String
    label: StringOptional
    description: StringOptional
    group_type: Enum
    members: RelationshipManager
    subscribers: RelationshipManager
    parent: RelatedNode
    children: RelationshipManager


class CoreMenu(CoreNode):
    namespace: String
    name: String
    label: StringOptional
    kind: StringOptional
    path: StringOptional
    description: StringOptional
    icon: StringOptional
    protected: Boolean
    order_weight: Integer
    required_permissions: ListAttributeOptional
    section: Enum
    parent: RelatedNode
    children: RelationshipManager


class CoreNodeTriggerMatch(CoreNode):
    trigger: RelatedNode


class CoreObjectComponentTemplate(CoreNode):
    template_name: String


class CoreObjectTemplate(CoreNode):
    template_name: String


class CoreProfile(CoreNode):
    profile_name: String
    profile_priority: IntegerOptional


class CoreResourcePool(CoreNode):
    name: String
    description: StringOptional


class CoreTaskTarget(CoreNode):
    pass


class CoreThread(CoreNode):
    label: StringOptional
    resolved: Boolean
    created_at: DateTimeOptional
    change: RelatedNode
    comments: RelationshipManager
    created_by: RelatedNode


class CoreTransformation(CoreNode):
    name: String
    label: StringOptional
    description: StringOptional
    timeout: Integer
    query: RelatedNode
    repository: RelatedNode
    tags: RelationshipManager


class CoreTriggerRule(CoreNode):
    name: String
    description: StringOptional
    active: Boolean
    branch_scope: Dropdown
    action: RelatedNode


class CoreValidator(CoreNode):
    label: StringOptional
    state: Enum
    conclusion: Enum
    completed_at: DateTimeOptional
    started_at: DateTimeOptional
    proposed_change: RelatedNode
    checks: RelationshipManager


class CoreWebhook(CoreNode):
    name: String
    event_type: Enum
    active: Boolean
    branch_scope: Dropdown
    node_kind: StringOptional
    description: StringOptional
    url: URL
    validate_certificates: BooleanOptional


class CoreWeightedPoolResource(CoreNode):
    allocation_weight: IntegerOptional


class LineageOwner(CoreNode):
    pass


class LineageSource(CoreNode):
    pass


class BuiltinTag(CoreNode):
    name: String
    description: StringOptional


class CoreAccount(LineageOwner, LineageSource, CoreGenericAccount):
    pass


class CoreAccountGroup(LineageOwner, LineageSource, CoreGroup):
    roles: RelationshipManager


class CoreAccountRole(CoreNode):
    name: String
    groups: RelationshipManager
    permissions: RelationshipManager


class CoreArtifact(CoreTaskTarget):
    name: String
    status: Enum
    content_type: Enum
    checksum: StringOptional
    storage_id: StringOptional
    parameters: JSONAttributeOptional
    object: RelatedNode
    definition: RelatedNode


class CoreArtifactCheck(CoreCheck):
    changed: BooleanOptional
    checksum: StringOptional
    artifact_id: StringOptional
    storage_id: StringOptional
    line_number: IntegerOptional


class CoreArtifactDefinition(CoreTaskTarget):
    name: String
    artifact_name: String
    description: StringOptional
    parameters: JSONAttribute
    content_type: Enum
    targets: RelatedNode
    transformation: RelatedNode


class CoreArtifactThread(CoreThread):
    artifact_id: StringOptional
    storage_id: StringOptional
    line_number: IntegerOptional


class CoreArtifactValidator(CoreValidator):
    definition: RelatedNode


class CoreChangeComment(CoreComment):
    change: RelatedNode


class CoreChangeThread(CoreThread):
    pass


class CoreCheckDefinition(CoreTaskTarget):
    name: String
    description: StringOptional
    file_path: String
    class_name: String
    timeout: Integer
    parameters: JSONAttributeOptional
    repository: RelatedNode
    query: RelatedNode
    targets: RelatedNode
    tags: RelationshipManager


class CoreCustomWebhook(CoreWebhook, CoreTaskTarget):
    shared_key: StringOptional
    transformation: RelatedNode


class CoreDataCheck(CoreCheck):
    conflicts: JSONAttribute
    keep_branch: Enum
    enriched_conflict_id: StringOptional


class CoreDataValidator(CoreValidator):
    pass


class CoreFileCheck(CoreCheck):
    files: ListAttributeOptional
    commit: StringOptional


class CoreFileThread(CoreThread):
    file: StringOptional
    commit: StringOptional
    line_number: IntegerOptional
    repository: RelatedNode


class CoreGeneratorAction(CoreAction):
    generator: RelatedNode


class CoreGeneratorAwareGroup(CoreGroup):
    pass


class CoreGeneratorCheck(CoreCheck):
    instance: String


class CoreGeneratorDefinition(CoreTaskTarget):
    name: String
    description: StringOptional
    parameters: JSONAttribute
    file_path: String
    class_name: String
    convert_query_response: BooleanOptional
    execute_in_proposed_change: BooleanOptional
    execute_after_merge: BooleanOptional
    query: RelatedNode
    repository: RelatedNode
    targets: RelatedNode


class CoreGeneratorGroup(CoreGroup):
    pass


class CoreGeneratorInstance(CoreTaskTarget):
    name: String
    status: Enum
    object: RelatedNode
    definition: RelatedNode


class CoreGeneratorValidator(CoreValidator):
    definition: RelatedNode


class CoreGlobalPermission(CoreBasePermission):
    action: Dropdown
    decision: Enum


class CoreGraphQLQuery(CoreNode):
    name: String
    description: StringOptional
    query: String
    variables: JSONAttributeOptional
    operations: ListAttributeOptional
    models: ListAttributeOptional
    depth: IntegerOptional
    height: IntegerOptional
    repository: RelatedNode
    tags: RelationshipManager


class CoreGraphQLQueryGroup(CoreGroup):
    parameters: JSONAttributeOptional
    query: RelatedNode


class CoreGroupAction(CoreAction):
    member_action: Dropdown
    group: RelatedNode


class CoreGroupTriggerRule(CoreTriggerRule):
    member_update: Dropdown
    group: RelatedNode


class CoreIPAddressPool(CoreResourcePool, LineageSource):
    default_address_type: String
    default_prefix_length: IntegerOptional
    resources: RelationshipManager
    ip_namespace: RelatedNode


class CoreIPPrefixPool(CoreResourcePool, LineageSource):
    default_prefix_length: IntegerOptional
    default_member_type: Enum
    default_prefix_type: StringOptional
    resources: RelationshipManager
    ip_namespace: RelatedNode


class CoreMenuItem(CoreMenu):
    pass


class CoreNodeTriggerAttributeMatch(CoreNodeTriggerMatch):
    attribute_name: String
    value: StringOptional
    value_previous: StringOptional
    value_match: Dropdown


class CoreNodeTriggerRelationshipMatch(CoreNodeTriggerMatch):
    relationship_name: String
    modification_type: Dropdown
    peer: StringOptional


class CoreNodeTriggerRule(CoreTriggerRule):
    node_kind: String
    mutation_action: Enum
    matches: RelationshipManager


class CoreNumberPool(CoreResourcePool, LineageSource):
    node: String
    node_attribute: String
    start_range: Integer
    end_range: Integer
    pool_type: Enum


class CoreObjectPermission(CoreBasePermission):
    namespace: String
    name: String
    action: Enum
    decision: Enum


class CoreObjectThread(CoreThread):
    object_path: String


class CorePasswordCredential(CoreCredential):
    username: StringOptional
    password: StringOptional


class CoreProposedChange(CoreTaskTarget):
    name: String
    description: StringOptional
    source_branch: String
    destination_branch: String
    state: Enum
    is_draft: Boolean
    total_comments: IntegerOptional
    approved_by: RelationshipManager
    rejected_by: RelationshipManager
    reviewers: RelationshipManager
    comments: RelationshipManager
    threads: RelationshipManager
    validations: RelationshipManager


class CoreReadOnlyRepository(LineageOwner, LineageSource, CoreGenericRepository, CoreTaskTarget):
    ref: String
    commit: StringOptional


class CoreRepository(LineageOwner, LineageSource, CoreGenericRepository, CoreTaskTarget):
    default_branch: String
    commit: StringOptional


class CoreRepositoryGroup(CoreGroup):
    content: Dropdown
    repository: RelatedNode


class CoreRepositoryValidator(CoreValidator):
    repository: RelatedNode


class CoreSchemaCheck(CoreCheck):
    conflicts: JSONAttribute
    enriched_conflict_id: StringOptional


class CoreSchemaValidator(CoreValidator):
    pass


class CoreStandardCheck(CoreCheck):
    pass


class CoreStandardGroup(CoreGroup):
    pass


class CoreStandardWebhook(CoreWebhook, CoreTaskTarget):
    shared_key: String


class CoreThreadComment(CoreComment):
    thread: RelatedNode


class CoreTransformJinja2(CoreTransformation):
    template_path: String


class CoreTransformPython(CoreTransformation):
    file_path: String
    class_name: String
    convert_query_response: BooleanOptional


class CoreUserValidator(CoreValidator):
    check_definition: RelatedNode
    repository: RelatedNode


class InternalAccountToken(CoreNode):
    name: StringOptional
    token: String
    expiration: DateTimeOptional
    account: RelatedNode


class InternalIPPrefixAvailable(BuiltinIPPrefix):
    pass


class InternalIPRangeAvailable(BuiltinIPAddress):
    last_address: IPHost


class InternalRefreshToken(CoreNode):
    expiration: DateTime
    account: RelatedNode


class IpamNamespace(BuiltinIPNamespace):
    default: BooleanOptional


# ---------------------------------------------
# SYNC
# ---------------------------------------------


class BuiltinIPAddressSync(CoreNodeSync):
    address: IPHost
    description: StringOptional
    ip_namespace: RelatedNodeSync
    ip_prefix: RelatedNodeSync


class BuiltinIPNamespaceSync(CoreNodeSync):
    name: String
    description: StringOptional
    ip_prefixes: RelationshipManagerSync
    ip_addresses: RelationshipManagerSync


class BuiltinIPPrefixSync(CoreNodeSync):
    prefix: IPNetwork
    description: StringOptional
    member_type: Dropdown
    is_pool: Boolean
    is_top_level: BooleanOptional
    utilization: IntegerOptional
    netmask: StringOptional
    hostmask: StringOptional
    network_address: StringOptional
    broadcast_address: StringOptional
    ip_namespace: RelatedNodeSync
    ip_addresses: RelationshipManagerSync
    resource_pool: RelationshipManagerSync
    parent: RelatedNodeSync
    children: RelationshipManagerSync


class CoreActionSync(CoreNodeSync):
    name: String
    description: StringOptional
    triggers: RelationshipManagerSync


class CoreArtifactTargetSync(CoreNodeSync):
    artifacts: RelationshipManagerSync


class CoreBasePermissionSync(CoreNodeSync):
    description: StringOptional
    identifier: StringOptional
    roles: RelationshipManagerSync


class CoreCheckSync(CoreNodeSync):
    name: StringOptional
    label: StringOptional
    origin: String
    kind: String
    message: StringOptional
    conclusion: Enum
    severity: Enum
    created_at: DateTimeOptional
    validator: RelatedNodeSync


class CoreCommentSync(CoreNodeSync):
    text: String
    created_at: DateTimeOptional
    created_by: RelatedNodeSync


class CoreCredentialSync(CoreNodeSync):
    name: String
    label: StringOptional
    description: StringOptional


class CoreFileObjectSync(CoreNodeSync):
    file_name: String
    checksum: String
    file_size: Integer
    file_type: String
    storage_id: String


class CoreGenericAccountSync(CoreNodeSync):
    name: String
    password: HashedPassword
    label: StringOptional
    description: StringOptional
    account_type: Enum
    status: Dropdown
    tokens: RelationshipManagerSync


class CoreGenericRepositorySync(CoreNodeSync):
    name: String
    description: StringOptional
    location: String
    internal_status: Dropdown
    operational_status: Dropdown
    sync_status: Dropdown
    credential: RelatedNodeSync
    tags: RelationshipManagerSync
    transformations: RelationshipManagerSync
    queries: RelationshipManagerSync
    checks: RelationshipManagerSync
    generators: RelationshipManagerSync
    groups_objects: RelationshipManagerSync


class CoreGroupSync(CoreNodeSync):
    name: String
    label: StringOptional
    description: StringOptional
    group_type: Enum
    members: RelationshipManagerSync
    subscribers: RelationshipManagerSync
    parent: RelatedNodeSync
    children: RelationshipManagerSync


class CoreMenuSync(CoreNodeSync):
    namespace: String
    name: String
    label: StringOptional
    kind: StringOptional
    path: StringOptional
    description: StringOptional
    icon: StringOptional
    protected: Boolean
    order_weight: Integer
    required_permissions: ListAttributeOptional
    section: Enum
    parent: RelatedNodeSync
    children: RelationshipManagerSync


class CoreNodeTriggerMatchSync(CoreNodeSync):
    trigger: RelatedNodeSync


class CoreObjectComponentTemplateSync(CoreNodeSync):
    template_name: String


class CoreObjectTemplateSync(CoreNodeSync):
    template_name: String


class CoreProfileSync(CoreNodeSync):
    profile_name: String
    profile_priority: IntegerOptional


class CoreResourcePoolSync(CoreNodeSync):
    name: String
    description: StringOptional


class CoreTaskTargetSync(CoreNodeSync):
    pass


class CoreThreadSync(CoreNodeSync):
    label: StringOptional
    resolved: Boolean
    created_at: DateTimeOptional
    change: RelatedNodeSync
    comments: RelationshipManagerSync
    created_by: RelatedNodeSync


class CoreTransformationSync(CoreNodeSync):
    name: String
    label: StringOptional
    description: StringOptional
    timeout: Integer
    query: RelatedNodeSync
    repository: RelatedNodeSync
    tags: RelationshipManagerSync


class CoreTriggerRuleSync(CoreNodeSync):
    name: String
    description: StringOptional
    active: Boolean
    branch_scope: Dropdown
    action: RelatedNodeSync


class CoreValidatorSync(CoreNodeSync):
    label: StringOptional
    state: Enum
    conclusion: Enum
    completed_at: DateTimeOptional
    started_at: DateTimeOptional
    proposed_change: RelatedNodeSync
    checks: RelationshipManagerSync


class CoreWebhookSync(CoreNodeSync):
    name: String
    event_type: Enum
    active: Boolean
    branch_scope: Dropdown
    node_kind: StringOptional
    description: StringOptional
    url: URL
    validate_certificates: BooleanOptional


class CoreWeightedPoolResourceSync(CoreNodeSync):
    allocation_weight: IntegerOptional


class LineageOwnerSync(CoreNodeSync):
    pass


class LineageSourceSync(CoreNodeSync):
    pass


class BuiltinTagSync(CoreNodeSync):
    name: String
    description: StringOptional


class CoreAccountSync(LineageOwnerSync, LineageSourceSync, CoreGenericAccountSync):
    pass


class CoreAccountGroupSync(LineageOwnerSync, LineageSourceSync, CoreGroupSync):
    roles: RelationshipManagerSync


class CoreAccountRoleSync(CoreNodeSync):
    name: String
    groups: RelationshipManagerSync
    permissions: RelationshipManagerSync


class CoreArtifactSync(CoreTaskTargetSync):
    name: String
    status: Enum
    content_type: Enum
    checksum: StringOptional
    storage_id: StringOptional
    parameters: JSONAttributeOptional
    object: RelatedNodeSync
    definition: RelatedNodeSync


class CoreArtifactCheckSync(CoreCheckSync):
    changed: BooleanOptional
    checksum: StringOptional
    artifact_id: StringOptional
    storage_id: StringOptional
    line_number: IntegerOptional


class CoreArtifactDefinitionSync(CoreTaskTargetSync):
    name: String
    artifact_name: String
    description: StringOptional
    parameters: JSONAttribute
    content_type: Enum
    targets: RelatedNodeSync
    transformation: RelatedNodeSync


class CoreArtifactThreadSync(CoreThreadSync):
    artifact_id: StringOptional
    storage_id: StringOptional
    line_number: IntegerOptional


class CoreArtifactValidatorSync(CoreValidatorSync):
    definition: RelatedNodeSync


class CoreChangeCommentSync(CoreCommentSync):
    change: RelatedNodeSync


class CoreChangeThreadSync(CoreThreadSync):
    pass


class CoreCheckDefinitionSync(CoreTaskTargetSync):
    name: String
    description: StringOptional
    file_path: String
    class_name: String
    timeout: Integer
    parameters: JSONAttributeOptional
    repository: RelatedNodeSync
    query: RelatedNodeSync
    targets: RelatedNodeSync
    tags: RelationshipManagerSync


class CoreCustomWebhookSync(CoreWebhookSync, CoreTaskTargetSync):
    shared_key: StringOptional
    transformation: RelatedNodeSync


class CoreDataCheckSync(CoreCheckSync):
    conflicts: JSONAttribute
    keep_branch: Enum
    enriched_conflict_id: StringOptional


class CoreDataValidatorSync(CoreValidatorSync):
    pass


class CoreFileCheckSync(CoreCheckSync):
    files: ListAttributeOptional
    commit: StringOptional


class CoreFileThreadSync(CoreThreadSync):
    file: StringOptional
    commit: StringOptional
    line_number: IntegerOptional
    repository: RelatedNodeSync


class CoreGeneratorActionSync(CoreActionSync):
    generator: RelatedNodeSync


class CoreGeneratorAwareGroupSync(CoreGroupSync):
    pass


class CoreGeneratorCheckSync(CoreCheckSync):
    instance: String


class CoreGeneratorDefinitionSync(CoreTaskTargetSync):
    name: String
    description: StringOptional
    parameters: JSONAttribute
    file_path: String
    class_name: String
    convert_query_response: BooleanOptional
    execute_in_proposed_change: BooleanOptional
    execute_after_merge: BooleanOptional
    query: RelatedNodeSync
    repository: RelatedNodeSync
    targets: RelatedNodeSync


class CoreGeneratorGroupSync(CoreGroupSync):
    pass


class CoreGeneratorInstanceSync(CoreTaskTargetSync):
    name: String
    status: Enum
    object: RelatedNodeSync
    definition: RelatedNodeSync


class CoreGeneratorValidatorSync(CoreValidatorSync):
    definition: RelatedNodeSync


class CoreGlobalPermissionSync(CoreBasePermissionSync):
    action: Dropdown
    decision: Enum


class CoreGraphQLQuerySync(CoreNodeSync):
    name: String
    description: StringOptional
    query: String
    variables: JSONAttributeOptional
    operations: ListAttributeOptional
    models: ListAttributeOptional
    depth: IntegerOptional
    height: IntegerOptional
    repository: RelatedNodeSync
    tags: RelationshipManagerSync


class CoreGraphQLQueryGroupSync(CoreGroupSync):
    parameters: JSONAttributeOptional
    query: RelatedNodeSync


class CoreGroupActionSync(CoreActionSync):
    member_action: Dropdown
    group: RelatedNodeSync


class CoreGroupTriggerRuleSync(CoreTriggerRuleSync):
    member_update: Dropdown
    group: RelatedNodeSync


class CoreIPAddressPoolSync(CoreResourcePoolSync, LineageSourceSync):
    default_address_type: String
    default_prefix_length: IntegerOptional
    resources: RelationshipManagerSync
    ip_namespace: RelatedNodeSync


class CoreIPPrefixPoolSync(CoreResourcePoolSync, LineageSourceSync):
    default_prefix_length: IntegerOptional
    default_member_type: Enum
    default_prefix_type: StringOptional
    resources: RelationshipManagerSync
    ip_namespace: RelatedNodeSync


class CoreMenuItemSync(CoreMenuSync):
    pass


class CoreNodeTriggerAttributeMatchSync(CoreNodeTriggerMatchSync):
    attribute_name: String
    value: StringOptional
    value_previous: StringOptional
    value_match: Dropdown


class CoreNodeTriggerRelationshipMatchSync(CoreNodeTriggerMatchSync):
    relationship_name: String
    modification_type: Dropdown
    peer: StringOptional


class CoreNodeTriggerRuleSync(CoreTriggerRuleSync):
    node_kind: String
    mutation_action: Enum
    matches: RelationshipManagerSync


class CoreNumberPoolSync(CoreResourcePoolSync, LineageSourceSync):
    node: String
    node_attribute: String
    start_range: Integer
    end_range: Integer
    pool_type: Enum


class CoreObjectPermissionSync(CoreBasePermissionSync):
    namespace: String
    name: String
    action: Enum
    decision: Enum


class CoreObjectThreadSync(CoreThreadSync):
    object_path: String


class CorePasswordCredentialSync(CoreCredentialSync):
    username: StringOptional
    password: StringOptional


class CoreProposedChangeSync(CoreTaskTargetSync):
    name: String
    description: StringOptional
    source_branch: String
    destination_branch: String
    state: Enum
    is_draft: Boolean
    total_comments: IntegerOptional
    approved_by: RelationshipManagerSync
    rejected_by: RelationshipManagerSync
    reviewers: RelationshipManagerSync
    comments: RelationshipManagerSync
    threads: RelationshipManagerSync
    validations: RelationshipManagerSync


class CoreReadOnlyRepositorySync(LineageOwnerSync, LineageSourceSync, CoreGenericRepositorySync, CoreTaskTargetSync):
    ref: String
    commit: StringOptional


class CoreRepositorySync(LineageOwnerSync, LineageSourceSync, CoreGenericRepositorySync, CoreTaskTargetSync):
    default_branch: String
    commit: StringOptional


class CoreRepositoryGroupSync(CoreGroupSync):
    content: Dropdown
    repository: RelatedNodeSync


class CoreRepositoryValidatorSync(CoreValidatorSync):
    repository: RelatedNodeSync


class CoreSchemaCheckSync(CoreCheckSync):
    conflicts: JSONAttribute
    enriched_conflict_id: StringOptional


class CoreSchemaValidatorSync(CoreValidatorSync):
    pass


class CoreStandardCheckSync(CoreCheckSync):
    pass


class CoreStandardGroupSync(CoreGroupSync):
    pass


class CoreStandardWebhookSync(CoreWebhookSync, CoreTaskTargetSync):
    shared_key: String


class CoreThreadCommentSync(CoreCommentSync):
    thread: RelatedNodeSync


class CoreTransformJinja2Sync(CoreTransformationSync):
    template_path: String


class CoreTransformPythonSync(CoreTransformationSync):
    file_path: String
    class_name: String
    convert_query_response: BooleanOptional


class CoreUserValidatorSync(CoreValidatorSync):
    check_definition: RelatedNodeSync
    repository: RelatedNodeSync


class InternalAccountTokenSync(CoreNodeSync):
    name: StringOptional
    token: String
    expiration: DateTimeOptional
    account: RelatedNodeSync


class InternalIPPrefixAvailableSync(BuiltinIPPrefixSync):
    pass


class InternalIPRangeAvailableSync(BuiltinIPAddressSync):
    last_address: IPHost


class InternalRefreshTokenSync(CoreNodeSync):
    expiration: DateTime
    account: RelatedNodeSync


class IpamNamespaceSync(BuiltinIPNamespaceSync):
    default: BooleanOptional



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType

from ..shared import GraphMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class SchemaRelationshipCleanupQuery(Query):
    name = "schema_relationship_cleanup"
    type: QueryType = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
        // ---------------------------------------------------------------
        // Find all "profiles" and "object_template" SchemaRelationship nodes,
        // plus any SchemaRelationship with no parent SchemaNode/SchemaGeneric
        // ---------------------------------------------------------------
        CALL {
            MATCH (sr:SchemaRelationship)-[:HAS_ATTRIBUTE]->(:Attribute {name: "name"})
                  -[:HAS_VALUE]->(av:AttributeValue)
            WHERE av.value IN ["profiles", "object_template"]
            RETURN sr
            UNION
            MATCH (sr:SchemaRelationship)
            WHERE NOT EXISTS {
                MATCH (sr)-[:IS_RELATED]-(:Relationship {name: "schema__node__relationships"})
            }
            RETURN sr
        }
        WITH DISTINCT sr AS sr_to_delete

        // ---------------------------------------------------------------
        // Collect any connected Relationship vertices for cleanup
        // ---------------------------------------------------------------
        OPTIONAL MATCH (rel_to_delete:Relationship)-[:IS_RELATED]-(sr_to_delete)

        // ---------------------------------------------------------------
        // Collect all Attributes and AttributeValues for cleanup
        // ---------------------------------------------------------------
        MATCH (sr_to_delete)-[:HAS_ATTRIBUTE]->(attribute_to_delete:Attribute)-[:HAS_VALUE]->(av_to_maybe_delete:AttributeValue)
        DETACH DELETE rel_to_delete, sr_to_delete, attribute_to_delete

        // ---------------------------------------------------------------
        // Delete orphaned AttributeValue nodes
        // ---------------------------------------------------------------
        WITH DISTINCT av_to_maybe_delete
        WHERE NOT EXISTS {
            MATCH (av_to_maybe_delete)-[]-()
        }
        DELETE av_to_maybe_delete
        """
        self.add_to_query(query)


class Migration058(GraphMigration):
    name: str = "058_remove_virtual_schema_relationships"
    queries: Sequence[type[Query]] = [SchemaRelationshipCleanupQuery]
    minimum_version: int = 57

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import ujson

if TYPE_CHECKING:
    from typing import BinaryIO


class MultipartBuilder:
    """Builds multipart form data payloads for GraphQL file uploads.

    This class implements the GraphQL Multipart Request Spec for uploading files via GraphQL mutations. The spec defines a standard way to send files
    alongside GraphQL operations using multipart/form-data.

    The payload structure follows the spec:
    - operations: JSON containing the GraphQL query and variables
    - map: JSON mapping file keys to variable paths
    - 0, 1, ...: The actual file contents

    Example payload:
        {
            "operations": '{"query": "mutation($file: Upload!) {...}", "variables": {"file": null}}',
            "map": '{"0": ["variables.file"]}',
            "0": (filename, file_content)
        }
    """

    @staticmethod
    def build_operations(query: str, variables: dict[str, Any]) -> str:
        """Build the operations JSON string.

        Args:
            query: The GraphQL query string.
            variables: The variables dict (file variable should be null).

        Returns:
            JSON string containing the query and variables.
        """
        return ujson.dumps({"query": query, "variables": variables})

    @staticmethod
    def build_file_map(file_key: str = "0", variable_path: str = "variables.file") -> str:
        """Build the file map JSON string.

        Args:
            file_key: The key used for the file in the multipart payload.
            variable_path: The path to the file variable in the GraphQL variables.

        Returns:
            JSON string mapping the file key to the variable path.
        """
        return ujson.dumps({file_key: [variable_path]})

    @staticmethod
    def build_payload(
        query: str,
        variables: dict[str, Any],
        file_content: BinaryIO | None = None,
        file_name: str = "upload",
    ) -> dict[str, Any]:
        """Build the complete multipart form data payload.

        Constructs the payload according to the GraphQL Multipart Request Spec. The returned dict can be passed directly to httpx as the `files`
        parameter.

        Args:
            query: The GraphQL query string containing $file: Upload! variable.
            variables: The variables dict. The 'file' key will be set to null.
            file_content: The file content as a file-like object (BinaryIO).
                          If None, only the operations and map will be included.
            file_name: The filename to use for the upload.

        Returns:
            A dict suitable for httpx's `files` parameter in a POST request.

        Example:
            >>> builder = MultipartBuilder()
            >>> payload = builder.build_payload(
            ...     query="mutation($file: Upload!) { upload(file: $file) { id } }",
            ...     variables={"other": "value"},
            ...     file_content=open("file.pdf", "rb"),
            ...     file_name="document.pdf",
            ... )
            >>> # payload can be passed to httpx.post(..., files=payload)
        """
        # Ensure file variable is null (spec requirement)
        variables = {**variables, "file": None}

        operations = MultipartBuilder.build_operations(query=query, variables=variables)
        file_map = MultipartBuilder.build_file_map()

        files: dict[str, Any] = {"operations": (None, operations), "map": (None, file_map)}

        if file_content is not None:
            # httpx streams from file-like objects automatically
            files["0"] = (file_name, file_content)

        return files



import hashlib
from typing import TYPE_CHECKING

from infrahub.core.constants import RelationshipCardinality
from infrahub.core.node import Node
from infrahub.core.schema import GenericSchema
from infrahub.core.schema.schema_branch import SchemaBranch

if TYPE_CHECKING:
    from infrahub.core.relationship import RelationshipManager


RESOURCE_POOL_LOCK_NAMESPACE = "resource_pool"
RELATIONSHIP_COUNT_LOCK_NAMESPACE = "relationship_count"


def _get_kinds_to_lock_on_object_mutation(kind: str, schema_branch: SchemaBranch) -> list[str]:
    """
    Return kinds for which we want to lock during creating / updating an object of a given schema node.
    Lock should be performed on schema kind and its generics having a uniqueness_constraint defined.
    If a generic uniqueness constraint is the same as the node schema one,
    it means node schema overrided this constraint, in which case we only need to lock on the generic.
    """

    node_schema = schema_branch.get(name=kind, duplicate=False)

    schema_uc = None
    kinds = []
    if node_schema.uniqueness_constraints:
        kinds.append(node_schema.kind)
        schema_uc = node_schema.uniqueness_constraints

    if isinstance(node_schema, GenericSchema):
        return kinds

    generics_kinds = node_schema.inherit_from

    node_schema_kind_removed = False
    for generic_kind in generics_kinds:
        generic_uc = schema_branch.get(name=generic_kind, duplicate=False).uniqueness_constraints
        if generic_uc:
            kinds.append(generic_kind)
            if not node_schema_kind_removed and generic_uc == schema_uc:
                # Check whether we should remove original schema kind as it simply overrides uniqueness_constraint
                # of a generic
                kinds.pop(0)
                node_schema_kind_removed = True
    return kinds


def _hash(value: str) -> str:
    # Do not use builtin `hash` for lock names as due to randomization results would differ between
    # different processes.
    return hashlib.sha256(value.encode()).hexdigest()


def get_lock_names_on_object_mutation(node: Node, schema_branch: SchemaBranch) -> list[str]:
    """
    Return lock names for object on which we want to avoid concurrent mutation (create/update).
    Lock names include kind, some generic kinds, resource pool ids, peer ids for cardinality one relationships,
    and values of attributes of corresponding uniqueness constraints.
    """

    lock_names: set[str] = set()

    # Check if node is using resource manager allocation via attributes
    for attr_name in node.get_schema().attribute_names:
        attribute = getattr(node, attr_name, None)
        if attribute is not None and getattr(attribute, "from_pool", None) and "id" in attribute.from_pool:
            lock_names.add(f"{RESOURCE_POOL_LOCK_NAMESPACE}.{attribute.from_pool['id']}")

    # Check if relationships allocate resources or have cardinality one constraint
    for rel_name in node._relationships:
        rel_manager: RelationshipManager = getattr(node, rel_name)
        for rel in rel_manager._relationships:
            if rel.from_pool and "id" in rel.from_pool:
                lock_names.add(f"{RESOURCE_POOL_LOCK_NAMESPACE}.{rel.from_pool['id']}")

            peer_id = rel.peer_id
            if not peer_id or not rel.schema.identifier:
                continue

            # Check if this node's relationship has cardinality one or max/min_count constraint
            # This prevents concurrent updates to the same node's constrained relationship
            if rel.schema.cardinality == RelationshipCardinality.ONE or rel.schema.max_count or rel.schema.min_count:
                lock_names.add(f"{RELATIONSHIP_COUNT_LOCK_NAMESPACE}.{rel.schema.identifier}.{node.id}")

            # Check if the peer has count constraints on the reverse relationship
            # This includes cardinality one, max_count, and min_count constraints
            peer_schema = schema_branch.get(name=rel.schema.peer, duplicate=False)
            peer_rel = peer_schema.get_relationship_by_identifier(id=rel.schema.identifier, raise_on_error=False)
            if peer_rel and (
                peer_rel.cardinality == RelationshipCardinality.ONE or peer_rel.max_count or peer_rel.min_count
            ):
                lock_names.add(f"{RELATIONSHIP_COUNT_LOCK_NAMESPACE}.{rel.schema.identifier}.{peer_id}")

    lock_kinds = _get_kinds_to_lock_on_object_mutation(node.get_kind(), schema_branch)
    for kind in lock_kinds:
        schema = schema_branch.get(name=kind, duplicate=False)
        ucs = schema.uniqueness_constraints
        if ucs is None:
            continue

        ucs_lock_names: list[str] = []
        uc_attributes_names = set()

        for uc in ucs:
            uc_attributes_values = []
            # Keep only attributes constraints
            for field_path in uc:
                # Some attributes may exist in different uniqueness constraints, we de-duplicate them
                if field_path in uc_attributes_names:
                    continue

                # Exclude relationships uniqueness constraints
                schema_path = schema.parse_schema_path(path=field_path, schema=schema_branch)
                if schema_path.related_schema is not None or schema_path.attribute_schema is None:
                    continue

                uc_attributes_names.add(field_path)
                attr = getattr(node, schema_path.attribute_schema.name, None)
                if attr is None or attr.value is None:
                    # `attr.value` being None corresponds to optional unique attribute.
                    # `attr` being None is not supposed to happen.
                    value_hashed = _hash("")
                else:
                    value_hashed = _hash(str(attr.value))

                uc_attributes_values.append(value_hashed)

            if uc_attributes_values:
                uc_lock_name = ".".join(uc_attributes_values)
                ucs_lock_names.append(uc_lock_name)

        if not ucs_lock_names:
            continue

        partial_lock_name = kind + "." + ".".join(ucs_lock_names)
        lock_names.add(build_object_lock_name(partial_lock_name))

    return sorted(lock_names)


def build_object_lock_name(name: str) -> str:
    return f"global.object.{name}"



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any, Iterable

if TYPE_CHECKING:
    from infrahub.core.query import QueryResult
    from infrahub.core.schema import SchemaAttributePath, SchemaAttributePathValue


class GroupedIndexKey:
    def __init__(self) -> None:
        self._keys: list[tuple[str, str]] = []

    def add_key(self, key: tuple[str, Any]) -> None:
        self._keys.append(key)

    def get_key(self) -> list[tuple[str, Any]]:
        return self._keys

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self.get_key() == other.get_key()

    def __hash__(self) -> int:
        to_hash = ".".join([f"{k[0]}-{k[1]}" for k in self._keys])
        return hash(to_hash)


class UniquenessQueryResultsIndex:
    def __init__(self, query_results: Iterable[QueryResult], exclude_node_ids: set[str] | None = None) -> None:
        self._relationship_index: dict[str, dict[str, set[str]]] = {}
        self._attribute_index: dict[str, dict[Any, set[str]]] = {}
        self._node_index: dict[str, dict[str, Any]] = {}
        self._all_node_ids: set[str] = set()
        for query_result in query_results:
            node_id = query_result.get_as_str("node_id")
            if not node_id or (exclude_node_ids and node_id in exclude_node_ids):
                continue
            if node_id not in self._node_index:
                self._node_index[node_id] = {}
            self._all_node_ids.add(node_id)
            relationship_identifier = query_result.get_as_str("relationship_identifier")
            attr_name = query_result.get_as_str("attr_name")
            attr_value = query_result.get_as_str("attr_value")
            if relationship_identifier:
                if relationship_identifier not in self._relationship_index:
                    self._relationship_index[relationship_identifier] = defaultdict(set)
                if attr_value and node_id:
                    self._relationship_index[relationship_identifier][attr_value].add(node_id)
                    self._node_index[node_id][relationship_identifier] = attr_value
            elif attr_name:
                if attr_name not in self._attribute_index:
                    self._attribute_index[attr_name] = defaultdict(set)
                if attr_value and node_id:
                    self._attribute_index[attr_name][attr_value].add(node_id)
                    self._node_index[node_id][attr_name] = attr_value

    def get_node_ids_for_value_group(self, path_value_group: list[SchemaAttributePathValue]) -> set[str]:
        matching_node_ids = self._all_node_ids.copy()
        for schema_attribute_path_value in path_value_group:
            if schema_attribute_path_value.value is None:
                return set()
            value = str(schema_attribute_path_value.value)
            if schema_attribute_path_value.relationship_schema:
                relationship_identifier = schema_attribute_path_value.relationship_schema.get_identifier()
                matching_node_ids &= self._relationship_index.get(relationship_identifier, {}).get(value, set())
            elif schema_attribute_path_value.attribute_schema:
                attribute_name = schema_attribute_path_value.attribute_schema.name
                matching_node_ids &= self._attribute_index.get(attribute_name, {}).get(value, set())
            if not matching_node_ids:
                return matching_node_ids
        return matching_node_ids

    def get_node_ids_for_path_group(self, path_group: list[SchemaAttributePath]) -> set[str]:
        node_ids_by_attr_name_and_value: dict[GroupedIndexKey, set[str]] = {}
        key_group = []
        for schema_attribute_path in path_group:
            if schema_attribute_path.relationship_schema:
                key_group.append(schema_attribute_path.relationship_schema.get_identifier())
            elif schema_attribute_path.attribute_schema:
                key_group.append(schema_attribute_path.attribute_schema.name)
            else:
                continue
        for node_id, attribute_details in self._node_index.items():
            node_includes_none = False
            grouped_key = GroupedIndexKey()
            for element_key in key_group:
                element_value = attribute_details.get(element_key)
                if element_value is None:
                    node_includes_none = True
                    break
                grouped_key.add_key((element_key, element_value))
            if node_includes_none:
                continue
            if grouped_key not in node_ids_by_attr_name_and_value:
                node_ids_by_attr_name_and_value[grouped_key] = set()
            node_ids_by_attr_name_and_value[grouped_key].add(node_id)
        node_ids_in_violation = set()
        for node_ids in node_ids_by_attr_name_and_value.values():
            if len(node_ids) > 1:
                node_ids_in_violation |= node_ids
        return node_ids_in_violation
